#!/usr/bin/env python
# coding=utf-8
"""QC Config objects

Module to store the different QC modules in ioos_qc

Attributes:
    tw (namedtuple): The TimeWindow namedtuple definition
"""
import io
import logging
import warnings
from pathlib import Path
from copy import deepcopy
from inspect import signature
from functools import partial
from typing import Union, List
from importlib import import_module
from dataclasses import dataclass, field
from collections import namedtuple, OrderedDict as odict

import numpy as np
from shapely.geometry import shape, GeometryCollection

from ioos_qc.results import CallResult, collect_results
from ioos_qc.utils import load_config_as_dict, dict_depth

L = logging.getLogger(__name__)  # noqa
ConfigTypes = Union[dict, odict, str, Path, io.StringIO]

tw = namedtuple('TimeWindow', ('starting', 'ending'), defaults=[None, None])


@dataclass(frozen=True)
class Context:
    window: tw = field(default_factory=tw)
    region: GeometryCollection = field(default=None)
    attrs: dict = field(default_factory=dict)

    def __eq__(self, other):
        if isinstance(other, Context):
            return self.window == other.window and self.region == other.region
        return False

    def __key__(self):
        return (
            self.window,
            getattr(self.region, 'wkb', None)
        )

    def __hash__(self):
        return hash(self.__key__())

    def __repr__(self):
        return f'<Context window={self.window} region={self.region}>'


@dataclass(frozen=True)
class Call:
    stream_id: str
    call: partial
    context: Context = field(default_factory=Context)
    attrs: dict = field(default_factory=dict)

    @property
    def window(self):
        return self.context.window

    @property
    def region(self):
        return self.context.region

    @property
    def func(self) -> str:
        return self.call.func

    @property
    def module(self) -> str:
        return self.func.__module__.replace('ioos_qc.', '')

    @property
    def method(self) -> str:
        return self.func.__name__

    @property
    def method_path(self) -> str:
        return f'{self.module}.{self.method}'

    @property
    def args(self) -> tuple:
        return self.call.args

    @property
    def kwargs(self) -> dict:
        return self.call.keywords

    def config(self) -> dict:
        return {
            self.module: {
                self.method: self.kwargs
            }
        }

    @property
    def is_aggregate(self) -> bool:
        return hasattr(self.func, 'aggregate') and self.func.aggregate is True

    def __key__(self):
        return (
            self.stream_id,
            self.context.__hash__(),
            self.module,
            self.method,
            self.args,
            tuple(self.kwargs.items())
        )

    def __hash__(self):
        return hash(self.__key__())

    def __eq__(self, other):
        if isinstance(other, Call):
            return self.__key__() == other.__key__()
        return NotImplemented

    def __repr__(self):
        ret = f'<Call stream_id={self.stream_id}'
        if self.context.window.starting:
            ret += f' starting={self.window.starting}'
        if self.context.window.ending:
            ret += f' ending={self.window.ending}'
        if self.context.region is not None:
            ret += ' region=True'
        ret += f' {self.module}.{self.method}({self.args}, {self.kwargs})>'
        return ret

    def run(self, **passedkwargs):
        results = []

        # Get our own copy of the kwargs object so we can change it
        testkwargs = deepcopy(passedkwargs)
        # Merges dicts
        testkwargs = odict({ **self.kwargs, **testkwargs })

        # Get the arguments that the test functions support
        sig = signature(self.func)
        valid_keywords = [
            p.name for p in sig.parameters.values()
            if p.kind == p.POSITIONAL_OR_KEYWORD
        ]
        testkwargs = {
            k: v for k, v in testkwargs.items()
            if k in valid_keywords
        }
        try:
            results.append(
                CallResult(
                    package=self.module,
                    test=self.method,
                    function=self.func,
                    results=self.func(**testkwargs)
                )
            )
        except Exception as e:
            L.error(f'Could not run "{self.module}.{self.method}: {e}')

        return results


def extract_calls(source) -> List[Call]:
    """
    Extracts call objects from a source object

    Args:
        source ([any]): The source of Call objects, this can be a:
            * Call object
            * list of Call objects
            * list of objects with the 'calls' attribute
            * NewConfig object
            * Object with the 'calls' attribute

    Returns:
        List[Call]: List of extracted Call objects

    """
    if isinstance(source, Call):
        return [source]
    elif isinstance(source, (tuple, list)):
        # list of Call objects
        calls = [ c for c in source if isinstance(c, Call) ]
        # list of objects with the 'calls' attribute
        [
            calls.extend([
                x for x in c.calls if isinstance(x, Call)
            ])
            for c in source if hasattr(c, 'calls')
        ]
        return calls
    elif isinstance(source, Config):
        # Config object
        return source.calls
    elif hasattr(source, 'calls'):
        # Object with the 'calls' attribute
        return source.calls
    return []


class Config:
    """ A class to load any ioos_qc configuration setup into a list of callable objects
    that will run quality checks. The resulting list of quality checks parsed from a config
    file can be appended and edited until they are ready to be run. On run the checks are
    consolidated into an efficient structure for indexing the dataset (stream) it is run against
    so things like subsetting by time and space only happen once for each test in the same Context.

    How the individual checks are collected is up to each individual Stream implementation, this
    class only pares various formats and versions of a config into a list of Call objects.
    """

    def __init__(self, source, version=None, default_stream_key='_stream'):
        """
        Args:
            source: The QC configuration representation in one of the following formats:
                python dict or odict
                JSON/YAML filepath (str or Path object)
                JSON/YAML str
                JSON/YAML StringIO
                netCDF4/xarray filepath
                netCDF4/xarray Dataset
                list of Call objects
        """
        # A fully encapsulated Call objects that are configured
        # There are later grouped by window/region to more efficiently process
        # groups of indexes and variables
        self._calls = []

        # If we are passed an object we can extract calls from do so
        # Else, process as a Config object
        extracted = extract_calls(source)
        if extracted:
            self._calls = extracted
        else:
            # Parse config based on version
            # Massage and return the correct type of config object depending on the input
            self.config = load_config_as_dict(source)
            if 'contexts' in self.config:
                # Return a list of ContextConfig
                for c in self.config['contexts']:
                    self._calls.extend(list(ContextConfig(c).calls))
            elif 'streams' in self.config:
                # Return a list with just one ContextConfig
                self._calls += list(ContextConfig(self.config).calls)
            elif dict_depth(self.config) >= 4:
                # This is a StreamConfig
                self._calls += list(ContextConfig(odict(streams=self.config)).calls)
            else:
                # This is a QcConfig
                self._calls += list(ContextConfig(odict(streams={default_stream_key: self.config})).calls)
                #raise ValueError("Can not add context to a QC Config object. Create it manually.")

    @property
    def contexts(self):
        """
        Group the calls into context groups and return them
        """
        contexts = {}
        for c in self._calls:
            if c.context in contexts:
                contexts[c.context].append(c)
            else:
                contexts[c.context] = [c]
        return contexts

    @property
    def stream_ids(self):
        """
        Return a list of unique stream_ids for the Config
        """

        streams = []
        stream_map = {}

        for c in self._calls:
            if c.stream_id not in stream_map:
                stream_map[c.stream_id] = True
                streams.append(c.stream_id)

        return streams

    @property
    def calls(self):
        return self._calls
        # Could need this in the future
        # return [
        #     c for c in self._calls
        #     if not hasattr(c.func, 'aggregate') or c.func.aggregate is False
        # ]

    @property
    def aggregate_calls(self):
        return [
            c for c in self._calls
            if hasattr(c.func, 'aggregate') and c.func.aggregate is True
        ]

    def has(self, stream_id : str, method: Union[callable, str]):
        if isinstance(method, str):
            for c in self._calls:
                if c.stream_id == stream_id and c.method_path == method:
                    return c
        elif isinstance(method, callable):
            for c in self._calls:
                if (c.stream_id == stream_id and
                    c.method == method.__module__ and
                    c.method == method.__name__
                   ):
                    return c
        return False

    def calls_by_stream_id(self, stream_id) -> List[Call]:
        calls = []
        for c in self._calls:
            if c.stream_id == stream_id:
                calls.append(c)
        return calls

    def add(self, source) -> None:
        """
        Adds a source of calls to this Config. See extract_calls for information on the
        types of objects accepted as the source parameter. The changes the internal .calls
        attribute and returns None.

        Args:
            source ([any]): The source of Call objects, this can be a:
                * Call object
                * list of Call objects
                * list of objects with the 'calls' attribute
                * Config object
                * Object with the 'calls' attribute
        """
        extracted = extract_calls(source)
        self._calls += extracted


class ContextConfig:
    """A collection of a Region, a TimeWindow and a list of StreamConfig objects

    Defines a set of quality checks to run against multiple input streams.
    This can include a region and a time window to subset any DataStreams by before running checks.

    region: None
    window:
        starting: 2020-01-01T00:00:00Z
        ending: 2020-04-01T00:00:00Z
    streams:
        variable1:    # stream_id
            qartod:   # StreamConfig
                location_test:
                    bbox: [-80, 40, -70, 60]
        variable2:    # stream_id
            qartod:   # StreamConfig
                gross_range_test:
                    suspect_span: [1, 11]
                    fail_span: [0, 12]

    Helper methods exist to run this check against a different inputs:
        * pandas.DataFrame, dask.DataFrame, netCDF4.Dataset, xarray.Dataset, ERDDAP URL

    Attributes:
        config (odict): dict representation of the parsed ContextConfig source
        region (GeometryCollection): A `shapely` object representing the valid geographic region
        window (namedtuple): A TimeWindow object representing the valid time period
        streams (odict): dict representation of the parsed StreamConfig objects
    """

    def __init__(self, source: ConfigTypes):
        self.config = load_config_as_dict(source)

        self._calls = []
        self.attrs = self.config.get('attrs', {})

        # Region
        self.region = None
        if 'region' in self.config:
            # Convert region to a GeometryCollection Shapely object.
            if isinstance(self.config['region'], GeometryCollection):
                self.region = self.config['region']
            elif self.config['region'] and 'features' in self.config['region']:
                # Feature based GeoJSON
                self.region = GeometryCollection([
                    shape(feature['geometry']) for feature in self.config['region']['features']
                ])
            elif self.config['region'] and 'geometry' in self.config['region']:
                # Geometry based GeoJSON
                self.region = GeometryCollection([
                    shape(self.config['region']['geometry'])
                ])
            else:
                L.warning('Ignoring region because it could not be parsed, is it valid GeoJSON?')

        # Window
        if 'window' in self.config and isinstance(self.config['window'], tw):
            self.window = self.config['window']
        elif 'window' in self.config:
            self.window = tw(**self.config['window'])
        else:
            self.window = tw()

        self.context = Context(
            window=self.window,
            region=self.region,
            attrs=self.attrs
        )

        # Extract each Call from the nested JSON
        """
        Calls
        This parses through available checks and selects the actual test functions
        to run, but doesn't actually run anything. It just sets up the object to be
        run later by iterating over the configs.
        """
        for stream_id, sc in self.config['streams'].items():

            for package, modules in sc.items():
                try:
                    testpackage = import_module('ioos_qc.{}'.format(package))
                except ImportError:
                    L.warning(f'No ioos_qc package "{package}" was found, skipping.')
                    continue

                for testname, kwargs in modules.items():
                    kwargs = kwargs or {}
                    if not hasattr(testpackage, testname):
                        L.warning(f'No ioos_qc method "{package}.{testname}" was found, skipping')
                        continue
                    else:
                        runfunc = getattr(testpackage, testname)

                    self._calls.append(
                        Call(
                            stream_id=stream_id,
                            context=self.context,
                            call=partial(runfunc, (), **kwargs),
                            attrs=getattr(sc, 'attrs', {})
                        )
                    )

    @property
    def calls(self):
        return self._calls

    def add(self, source) -> None:
        """
        Adds a source of calls to this ContextConfig. See extract_calls for information on the
        types of objects accepted as the source parameter. The changes the internal .calls
        attribute and returns None.

        Args:
            source ([any]): The source of Call objects, this can be a:
                * Call object
                * list of Call objects
                * list of objects with the 'calls' attribute
                * Config object
                * Object with the 'calls' attribute
        """
        extracted = extract_calls(source)
        self._calls.extend([ e for e in extracted if e.context == self.context ])

    def __str__(self):
        # sc = list(self.streams.keys())
        return (
            f"<ContextConfig "
            f"calls={len(self._calls)} "
            f"region={self.region is not None} "
            f"window={self.window.starting is not None or self.window.ending is not None}"
            ">"
        )

    def __repr__(self):
        return self.__str__()


class QcConfig(Config):
    def __init__(self, source, default_stream_key='_stream'):
        """
        A Config objects with no concept of a Stream ID. Typically used when running QC on a single
        stream. This just sets up a stream with the name passed in as the "default_stream_key"
        parameter.

        Args:
            source: The QC configuration representation in one of the following formats:
                python dict or odict
                JSON/YAML filepath (str or Path object)
                JSON/YAML str
                JSON/YAML StringIO
                netCDF4/xarray filepath
                netCDF4/xarray Dataset
                list of Call objects
            default_stream_key: The internal name of the stream, defaults to "_stream"
        """
        warnings.warn(
            "The QcConfig object is deprecated, please use Config directly",
            DeprecationWarning
        )
        self._default_stream_key = default_stream_key
        super().__init__(source, default_stream_key=default_stream_key)

    def run(self, **passedkwargs):
        from ioos_qc.streams import NumpyStream
        # Cleanup kwarg names
        passedkwargs['time'] = passedkwargs.pop('tinp', None)
        passedkwargs['z'] = passedkwargs.pop('zinp', None)
        # Convert input to numpy arrays which is required for NumpySteam
        for k in ['inp', 'time', 'z', 'lat', 'lon', 'geom']:
            if k not in passedkwargs or passedkwargs[k] is None:
                continue
            if not isinstance(passedkwargs[k], np.ndarray):
                passedkwargs[k] = np.array(passedkwargs[k])
        # Run the checks
        np_stream = NumpyStream(**passedkwargs)
        # Collect the results
        results = collect_results(np_stream.run(self), how='dict')
        # Strip out the default_stream_key
        return results[self._default_stream_key]


class NcQcConfig(Config):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            (
                'The NcQcConfig object has been replaced by ioos_qc.config.Config '
                'and ioos_qc.streams.XarrayStream'
            )
        )



#!python
from ioos_qc.config_creator import (
    CreatorConfig,
    QcConfigCreator,
    QcVariableConfig,
    QC_CONFIG_CREATOR_SCHEMA,
    VARIABLE_CONFIG_SCHEMA
)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'creator_config_file',
        type=str,
        help='path to creator config file (e.g. creator_config.json)'
    )
    parser.add_argument(
        'variable_config_file',
        type=str,
        help='path to variable config file (e.g. air_config.json)'
    )
    args = parser.parse_args()

    creator_config = CreatorConfig(args.creator_config_file, QC_CONFIG_CREATOR_SCHEMA)
    variable_config = QcVariableConfig(args.variable_config_file, VARIABLE_CONFIG_SCHEMA)
    qc = QcConfigCreator(creator_config)
    qc.create_config(variable_config)



#!/usr/bin/env python
# coding=utf-8
"""Deprecated module. Consider using ARGO instead."""
import warnings

from ioos_qc import argo


def pressure_check(inp):
    warnings.warn("gliders.pressure_check has been replaced by argo.pressure_increasing_test", DeprecationWarning)
    return argo.pressure_increasing_test(inp)



#!/usr/bin/env python
# coding=utf-8
import logging
from collections import defaultdict
from collections import OrderedDict as odict

import numpy as np
import pandas as pd
import xarray as xr
from xarray.core.indexing import remap_label_indexers

from ioos_qc.config import Config
from ioos_qc.utils import mapdates
from ioos_qc.results import ContextResult

L = logging.getLogger(__name__)  # noqa


class BaseStream:
    """Each stream should define how to return a list of datastreams along with their time and depth association.
    Each of these streams will passed through quality control configurations and returned back to it. Each stream
    needs to also define what to do with the resulting results (how to store them.)"""

    def __init__(self, *args, **kwargs):
        """
        df: the dataframe
        """
        pass

    def time(self):
        """Return the time array from the source dataset. This is useful when plotting QC results."""
        pass

    def data(self, stream_id):
        """Return the data array from the source dataset based on stream_id. This is useful when
        plotting QC results."""
        pass

    def run(self, config : Config):
        """Iterate over the configs, splitting the streams up by geographic and time window
        before applying the individual config using QcConfig.run(). Store results for future usage.
        """
        pass


class PandasStream:

    def __init__(self, df, time=None, z=None, lat=None, lon=None, geom=None):
        """
        df: the dataframe
        time: the column to use for time
        z: the column to use for depth
        lat: the column to use for latitude, this or geom is required if using regional subsets
        lon: the column to use for longitude, this or geom is required if using regional subsets
        geom: the column containing the geometry, this or lat and lon are required if using regional subsets
        """
        self.df = df
        self.time_column = time or 'time'
        self.z_column = z or 'z'
        self.lat_column = lat or 'lat'
        self.lon_column = lon or 'lon'
        self.geom_column = geom or 'geom'

        axis_columns = [
            self.time_column,
            self.z_column,
            self.lat_column,
            self.lon_column,
            self.geom_column
        ]
        self.axis_columns = [ x for x in axis_columns if x in df ]

    def time(self):
        return self.df[self.time_column]

    def data(self, stream_id):
        return self.df[stream_id]

    def run(self, config : Config):

        for context, calls in config.contexts.items():

            # Subset first by the stream id in each call
            stream_ids = []
            for call in calls:
                if call.stream_id not in self.df:
                    L.warning(f'{call.stream_id} is not a column in the dataframe, skipping')
                    continue
                stream_ids.append(call.stream_id)
            subset = self.df.loc[:, list(set(stream_ids + self.axis_columns))]

            if context.region:
                # TODO: yeah this does nothing right now
                # Figure out if this is a geopandas DataFrame already. If not, create one using
                # the specified lat_column and lon_column attributes in the constructor
                # if self.geom_column not in subset:
                #     subset = gpd.DataFrame(subset)
                #     subset[self.geom_column] = 'wut'
                # subset = subset[[ subset[self.geom_column].within(context.region) ]]
                pass

            if context.window.starting is not None or context.window.ending is not None:
                if self.time_column in self.axis_columns:
                    if context.window.starting:
                        subset = subset.loc[subset[self.time_column] >= context.window.starting, :]
                    if context.window.ending:
                        subset = subset.loc[subset[self.time_column] < context.window.ending, :]
                else:
                    L.warning(f'Skipping window subset, {self.time_column} not in columns')
                    pass

            # This is a boolean array of what was subset and tested based on the initial data feed
            # Take the index of the subset and set those to true
            subset_indexes = pd.Series(0, index=self.df.index, dtype='bool')
            subset_indexes.iloc[subset.index] = True

            # The source is subset, now the resulting rows need to be tested
            # Put together the static inputs that were subset for this config
            subset_kwargs = {}
            if self.time_column in self.axis_columns:
                subset_kwargs['tinp'] = subset.loc[:, self.time_column]
            if self.z_column in self.axis_columns:
                subset_kwargs['zinp'] = subset.loc[:, self.z_column]
            if self.lon_column in self.axis_columns:
                subset_kwargs['lon'] = subset.loc[:, self.lon_column]
            if self.lat_column in self.axis_columns:
                subset_kwargs['lat'] = subset.loc[:, self.lat_column]

            # Perform the "run" function on each Call
            for call in calls:

                # if call.is_aggregate:
                #     # We compute aggregates using the results
                #     continue

                if call.stream_id not in subset:
                    L.warning(f'{call.stream_id} not a column in the input dataframe, skipping')
                    continue

                data_input = subset.loc[:, call.stream_id]

                # This evaluates the generator test results
                run_result = list(call.run(
                    inp=data_input,
                    **subset_kwargs
                ))

                yield ContextResult(
                    results=run_result,
                    stream_id=call.stream_id,
                    subset_indexes=subset_indexes.values,
                    data=data_input.values,
                    tinp=subset_kwargs.get('tinp', pd.Series(dtype='datetime64[ns]')).values,
                    zinp=subset_kwargs.get('zinp', pd.Series(dtype='float64')).values,
                    lat=subset_kwargs.get('lat', pd.Series(dtype='float64')).values,
                    lon=subset_kwargs.get('lon', pd.Series(dtype='float64')).values,
                )


class NumpyStream:

    def __init__(self, inp=None, time=None, z=None, lat=None, lon=None, geom=None):
        """
        inp: a numpy array or a dictionary of numpy arrays where the keys are the stream ids
        time: numpy array of date-like objects.
        z: numpy array of z
        lat: numpy array of latitude, this or geom is required if using regional subsets
        lon: numpy array of longitude, this or geom is required if using regional subsets
        geom: numpy array of geometry, this or lat and lon are required if using regional subsets
        """
        self.inp = inp
        try:
            assert time is not None
            self.tinp = pd.DatetimeIndex(mapdates(time))
        except BaseException:
            self.tinp = time
        self.zinp = z
        self.lat = lat
        self.lon = lon
        self.geom = geom

    def time(self):
        return self.tinp

    def data(self, stream_id=None):
        return self.inp

    def run(self, config: Config):

        for context, calls in config.contexts.items():

            # This is a boolean array of what was subset and tested based on the initial data feed
            # Take the index of the subset and set those to true
            subset_indexes = np.full_like(self.inp, 1, dtype=bool)

            if context.region:
                # TODO: yeah this does nothing right now
                # Subset against the passed in lat/lons arrays in passedkwargs
                if self.lat is not None and self.lon is not None:
                    pass
                else:
                    L.warning('Skipping region subset, "lat" and "lon" must be passed into NumpySource')

            if context.window.starting is not None or context.window.ending is not None:
                if self.tinp is not None:
                    if context.window.starting:
                        subset_indexes = (subset_indexes) & (self.tinp >= context.window.starting)
                    if context.window.ending:
                        subset_indexes = (subset_indexes) & (self.tinp < context.window.ending)
                else:
                    L.warning('Skipping window subset, "time" array must be passed into "run"')
                    pass

            subset_kwargs = {}
            if self.tinp is not None:
                subset_kwargs['tinp'] = self.tinp[subset_indexes]
            if self.zinp is not None:
                subset_kwargs['zinp'] = self.zinp[subset_indexes]
            if self.lon is not None:
                subset_kwargs['lon'] = self.lon[subset_indexes]
            if self.lat is not None:
                subset_kwargs['lat'] = self.lat[subset_indexes]

            for call in calls:

                # If the input was passed in the config.
                # This is here for backwards compatibility and doesn't support
                # being a different size than what the subset/context size is.
                # Pass in values in the config should be deprecated in the future!
                if self.inp is None and 'inp' in call.kwargs:
                    self.inp = np.array(call.kwargs['inp'])
                    subset_indexes = np.full_like(self.inp, 1, dtype=bool)

                # Support more than one named inp, but fall back to a single
                if isinstance(self.inp, np.ndarray):
                    runinput = self.inp
                elif isinstance(self.inp, dict):
                    if call.stream_id in self.inp:
                        runinput = self.inp[call.stream_id]
                    else:
                        L.warning(f'{call.stream_id} not in input dict, skipping')
                        continue
                else:
                    L.error(f"Input is not a dict or np.ndarray, skipping {call.stream_id}")
                    continue

                # Slicing with [True] changes the shape of an array so always re-shape. That
                # will happen when the input array is of size 1. Corner case but still need to
                # handle it here.
                original_shape = runinput.shape
                data_input = runinput[subset_indexes].reshape(original_shape)

                # This evaluates the generator test results
                run_result = list(call.run(
                    inp=data_input,
                    **subset_kwargs
                ))

                yield ContextResult(
                    results=run_result,
                    stream_id=call.stream_id,
                    subset_indexes=subset_indexes,
                    data=data_input,
                    tinp=subset_kwargs.get('tinp', pd.Series(dtype='datetime64[ns]')).values,
                    zinp=subset_kwargs.get('zinp', pd.Series(dtype='float64').values),
                    lat=subset_kwargs.get('lat', pd.Series(dtype='float64').values),
                    lon=subset_kwargs.get('lon', pd.Series(dtype='float64').values),
                )


class NetcdfStream:

    def __init__(self, path_or_ncd, time=None, z=None, lat=None, lon=None, geom=None):
        self.path_or_ncd = path_or_ncd

        self.time_var = time or 'time'
        self.z_var = z or 'z'
        self.lat_var = lat or 'lat'
        self.lon_var = lon or 'lon'

    def time(self):
        do_close, ds = self._open()
        tdata = ds.variables[self.time_var]
        if do_close is True:
            ds.close()
        return tdata

    def data(self, stream_id):
        do_close, ds = self._open()
        vdata = ds.variables[stream_id]
        if do_close is True:
            ds.close()
        return vdata

    def _open(self):
        if isinstance(self.path_or_ncd, str):
            do_close = True
            ds = xr.open_dataset(self.path_or_ncd, decode_cf=False)
        else:
            do_close = False
            ds = self.path_or_ncd

        return do_close, ds

    def run(self, config: Config):

        do_close, ds = self._open()

        stream_ids = []
        for context, calls in config.contexts.items():
            for call in calls:
                if call.stream_id not in ds.variables:
                    L.warning(f'{call.stream_id} is not a variable in the netCDF dataset, skipping')
                    continue
                stream_ids.append(call.stream_id)

        # Find any var specific kwargs to pass onto the run
        varkwargs = { 'inp': {} }
        if self.time_var in ds.variables:
            varkwargs['time'] = pd.DatetimeIndex(mapdates(ds.variables[self.time_var].values))
        if self.z_var in ds.variables:
            varkwargs['z'] = ds.variables[self.z_var].values
        if self.lat_var in ds.variables:
            varkwargs['lat'] = ds.variables[self.lat_var].values
        if self.lon_var in ds.variables:
            varkwargs['lon'] = ds.variables[self.lon_var].values

        # Now populate the `inp` dict for each valid data stream
        for s in stream_ids:
            if s in ds.variables:
                varkwargs['inp'][s] = ds.variables[s].values

        if do_close is True:
            ds.close()

        ns = NumpyStream(**varkwargs)
        yield from ns.run(config)


class XarrayStream:

    def __init__(self, path_or_ncd, time=None, z=None, lat=None, lon=None):
        self.path_or_ncd = path_or_ncd

        self.time_var = time or 'time'
        self.z_var = z or 'z'
        self.lat_var = lat or 'lat'
        self.lon_var = lon or 'lon'

    def time(self):
        do_close, ds = self._open()
        tdata = ds[self.time_var].values
        if do_close is True:
            ds.close()
        return tdata

    def data(self, stream_id):
        do_close, ds = self._open()
        vdata = ds[stream_id].values
        if do_close is True:
            ds.close()
        return vdata

    def _open(self):
        if isinstance(self.path_or_ncd, str):
            do_close = True
            ds = xr.open_dataset(
                self.path_or_ncd,
                decode_cf=True,
                decode_coords=True,
                decode_times=True,
                mask_and_scale=True
            )
        else:
            do_close = False
            ds = self.path_or_ncd

        return do_close, ds

    def run(self, config: Config):

        # Magic for nested key generation
        # https://stackoverflow.com/a/27809959
        results = defaultdict(lambda: defaultdict(odict))

        do_close, ds = self._open()

        for context, calls in config.contexts.items():

            for call in calls:

                # Find any var specific kwargs to pass onto the run
                if call.stream_id not in ds.variables:
                    L.warning(f'{call.stream_id} is not a variable in the xarray dataset, skipping')
                    continue

                # Because the variables could have different dimensions
                # we calculate the coordinates and subset for each
                # This is xarray style subsetting, so will look something like:
                # {
                #     'time': slice(datetime.datetime(2020, 1, 1, 0, 0), datetime.datetime(2020, 4, 1, 0, 0), None)
                # }
                label_indexes = {}
                subset_kwargs = {}

                # Region subset
                # TODO: yeah this does nothing right now
                # Subset against the passed in lat/lons variable keys
                # and build up the subset dict to apply later

                # Time subset
                if self.time_var in ds[call.stream_id].coords:
                    if context.window.starting and context.window.ending:
                        label_indexes[self.time_var] = slice(context.window.starting, context.window.ending)

                subset_stream = ds[call.stream_id].sel(**label_indexes)

                if self.time_var in subset_stream.coords:
                    # Already subset with the stream, best case. Good netCDF file.
                    subset_kwargs['tinp'] = subset_stream.coords[self.time_var].values
                elif self.time_var in ds.variables and ds[self.time_var].dims == ds[call.stream_id].dims:
                    # Same dimensions as the stream, so use the same subset
                    subset_kwargs['tinp'] = ds[self.time_var].sel(**label_indexes).values
                elif self.time_var in ds.variables and ds[self.time_var].size == ds[call.stream_id].size:
                    # Not specifically connected, but hey, the user asked for it
                    subset_kwargs['tinp'] = ds[self.time_var].sel(**label_indexes).values

                if self.z_var in subset_stream.coords:
                    # Already subset with the stream, best case. Good netCDF file.
                    subset_kwargs['zinp'] = subset_stream.coords[self.z_var].values
                elif self.z_var in ds.variables and ds[self.z_var].dims == ds[call.stream_id].dims:
                    # Same dimensions as the stream, so use the same subset
                    subset_kwargs['zinp'] = ds[self.z_var].sel(**label_indexes).values
                elif self.z_var in ds.variables and ds[self.z_var].size == ds[call.stream_id].size:
                    # Not specifically connected, but hey, the user asked for it
                    subset_kwargs['zinp'] = ds[self.z_var].sel(**label_indexes).values

                if self.lat_var in subset_stream.coords:
                    # Already subset with the stream, best case. Good netCDF file.
                    subset_kwargs['lat'] = subset_stream.coords[self.lat_var].values
                elif self.lat_var in ds.variables and ds[self.lat_var].dims == ds[call.stream_id].dims:
                    # Same dimensions as the stream, so use the same subset
                    subset_kwargs['lat'] = ds[self.lat_var].sel(**label_indexes).values
                elif self.lat_var in ds.variables and ds[self.lat_var].size == ds[call.stream_id].size:
                    # Not specifically connected, but hey, the user asked for it
                    subset_kwargs['lat'] = ds[self.lat_var].sel(**label_indexes).values

                if self.lon_var in subset_stream.coords:
                    # Already subset with the stream, best case. Good netCDF file.
                    subset_kwargs['lon'] = subset_stream.coords[self.lon_var].values
                elif self.lon_var in ds.variables and ds[self.lon_var].dims == ds[call.stream_id].dims:
                    # Same dimensions as the stream, so use the same subset
                    subset_kwargs['lon'] = ds[self.lon_var].sel(**label_indexes).values
                elif self.lon_var in ds.variables and ds[self.lon_var].size == ds[call.stream_id].size:
                    # Not specifically connected, but hey, the user asked for it
                    subset_kwargs['lon'] = ds[self.lon_var].sel(**label_indexes).values

                data_input = subset_stream.values
                run_result = call.run(
                    **subset_kwargs,
                    **dict(inp=data_input)
                )

                # Here we turn the labeled xarray indexes into boolean index arrays that numpy
                # can use to subset a basic array. This takes each labeled index, converts it to
                # its integer index representation (label -> integers) and then matches the keys
                # on each label with the dimension of the data variable. This result should be
                # able to be used on the original data feed AS IS using a direct subset notation
                # data[subset_indexes]. I'm pretty sure this works and if it doesn't blame my cat.
                # We start by subsetting nothing
                subset_indexes = np.full_like(ds[call.stream_id].values, 0, dtype=bool)
                int_indexes, _ = remap_label_indexers(ds[call.stream_id], label_indexes)
                # Initial slicer will select everything. This selects all values in a dimension
                # if there are no labeled indexes for it.
                slicers = [ slice(None) for x in range(ds[call.stream_id].ndim) ]
                for index_key, index_value in int_indexes.items():
                    if index_key in ds[call.stream_id].dims:
                        slicers[ds[call.stream_id].dims.index(index_key)] = index_value
                # We started with an empty subset_indexes, not set to True what we actually subset
                # using the labeled dimensions.

                # Casting to a tuple to handle a numpy deprecation:
                # FutureWarning: Using a non-tuple sequence for multidimensional indexing is
                # deprecated; use `arr[tuple(seq)]` instead of `arr[seq]`. In the future this will
                # be interpreted as an array index, `arr[np.array(seq)]`, which will result either
                # in an error or a different result.
                subset_indexes[tuple(slicers)] = True

                yield ContextResult(
                    results=run_result,
                    stream_id=call.stream_id,
                    subset_indexes=subset_indexes,
                    data=data_input,
                    tinp=subset_kwargs.get('tinp', pd.Series(dtype='datetime64[ns]').values),
                    zinp=subset_kwargs.get('zinp', pd.Series(dtype='float64').values),
                    lat=subset_kwargs.get('lat', pd.Series(dtype='float64').values),
                    lon=subset_kwargs.get('lon', pd.Series(dtype='float64').values),
                )

        if do_close is True:
            ds.close()

        return results



#!/usr/bin/env python
# coding=utf-8
import io
import logging
import simplejson as json
from typing import Any, Union
from numbers import Real
from pyproj import Geod
from pathlib import Path
from datetime import date, datetime
from collections import OrderedDict as odict
from collections.abc import Mapping

import numpy as np
import pandas as pd
import xarray as xr
import geojson
from ruamel.yaml import YAML

N = Real
L = logging.getLogger(__name__)  # noqa


def add_flag_metadata(**kwargs):
    def wrapper(func : callable):
        for k, v in kwargs.items():
            setattr(func, k, v)
        return func
    return wrapper


def openf(p, **kwargs):
    """ Helper to allow one-line-lambdas to read file contents
    """
    with open(p, **kwargs) as f:
        return f.read()


def load_config_from_xarray(source):
    """Load an xarray dataset as a config dict
    """

    to_close = False
    if not isinstance(source, xr.Dataset):
        source = xr.open_dataset(source, decode_cf=False)
        to_close = True

    # If a global attribute exists, load as a YAML or JSON string and
    # ignore any config at the variable level
    if 'ioos_qc_config' in source.attrs:
        L.info("Using global attribute ioos_qc_config for QC config")
        return load_config_as_dict(source.attrs['ioos_qc_config'])

    # Iterate over variables and construct a config object
    y = odict()
    source = source.filter_by_attrs(
        ioos_qc_module=lambda x: x is not None,
        ioos_qc_test=lambda x: x is not None,
        ioos_qc_config=lambda x: x is not None,
        ioos_qc_target=lambda x: x is not None,
    )
    for dv in source.variables:
        if dv in source.dims:
            continue
        vobj = source[dv]

        # Because a data variables can have more than one check
        # associated with it we need to merge any existing configs
        # for this variable
        newdict = odict({
            vobj.ioos_qc_module: odict({
                vobj.ioos_qc_test: odict(json.loads(vobj.ioos_qc_config))
            })
        })
        merged = dict_update(
            y.get(vobj.ioos_qc_target, {}),
            newdict
        )
        y[vobj.ioos_qc_target] = merged

    # If we opened this xarray dataset from a file we should close it
    if to_close is True:
        source.close()

    return y


def load_config_as_dict(source : Union[str, dict, odict, Path, io.StringIO]
                        ) -> odict:
    """Load an object as a config dict. The source can be a dict, odict,
    YAML string, JSON string, a StringIO, or a file path to a valid YAML or JSON file.
    """
    yaml = YAML(typ='safe')
    if isinstance(source, odict):
        return source
    elif isinstance(source, dict):
        return odict(source)
    elif isinstance(source, xr.Dataset):
        return load_config_from_xarray(source)
    elif isinstance(source, (str, Path)):
        source = str(source)

        # Try to load as YAML, then JSON, then file path
        load_funcs = [
            lambda x: odict(yaml.load(x)),
            lambda x: odict(json.loads(x)),
            lambda x: load_config_from_xarray(x),
            lambda x: odict(yaml.load(openf(x))),  # noqa
            lambda x: odict(json.loads(openf(x))),  # noqa
        ]
        for lf in load_funcs:
            try:
                return lf(source)
            except BaseException:
                continue

    elif isinstance(source, io.StringIO):
        # Try to load as YAML, then JSON, then file path
        load_funcs = [
            lambda x: odict(yaml.load(x)),
            lambda x: odict(json.load(x)),
        ]
        for lf in load_funcs:
            try:
                return lf(source.getvalue())
            except BaseException:
                continue

    raise ValueError('Config source is not valid!')


def isfixedlength(lst : Union[list, tuple],
                  length : int
                  ) -> bool:
    if not isinstance(lst, (list, tuple)):
        raise ValueError('Required: list/tuple, Got: {}'.format(type(lst)))

    if len(lst) != length:
        raise ValueError(
            'Incorrect list/tuple length for {}. Required: {}, Got: {}'.format(
                lst,
                length,
                len(lst)
            )
        )

    return True


def isnan(v : Any) -> bool:
    return (
        v is None or
        v is np.nan or
        v is np.ma.masked
    )


def mapdates(dates):
    if hasattr(dates, 'dtype') and hasattr(dates.dtype, 'tz'):
        # pandas time objects with a datetime component, remove the timezone
        return dates.dt.tz_localize(None).astype('datetime64[ns]').to_numpy()
    elif hasattr(dates, 'dtype') and hasattr(dates, 'to_numpy'):
        # pandas time objects without a datetime component
        return dates.to_numpy().astype('datetime64[ns]')
    elif hasattr(dates, 'dtype') and np.issubdtype(dates.dtype, np.datetime64):
        # numpy datetime objects
        return dates.astype('datetime64[ns]')
    else:
        try:
            # Finally try unix epoch seconds
            return pd.to_datetime(dates, unit='s').values.astype('datetime64[ns]')
        except Exception:
            # strings work here but we don't advertise that
            return np.array(dates, dtype='datetime64[ns]')


def check_timestamps(times : np.ndarray,
                     max_time_interval : N = None
                     ) -> bool:
    """Sanity checks for timestamp arrays

    Checks that the times supplied are in monotonically increasing
    chronological order, and optionally that time intervals between
    measurements do not exceed a value `max_time_interval`.  Note that this is
    not a QARTOD test, but rather a utility test to make sure times are in the
    proper order and optionally do not have large gaps prior to processing the
    data.

    Args:
        times: Input array of timestamps
        max_time_interval: The interval between values should not exceed this
            value. [optional]
    """

    time_diff = np.diff(times)
    sort_diff = np.diff(sorted(times))
    # Check if there are differences between sorted and unsorted, and then
    # see if if there are any duplicate times.  Then check that none of the
    # diffs exceeds the sorted time.
    zero = np.array(0, dtype=time_diff.dtype)
    if not np.array_equal(time_diff, sort_diff) or np.any(sort_diff == zero):
        return False
    elif (max_time_interval is not None and
          np.any(sort_diff > max_time_interval)):
        return False
    else:
        return True


def dict_update(d : Mapping, u : Mapping) -> Mapping:
    # http://stackoverflow.com/a/3233356
    for k, v in u.items():
        if isinstance(d, Mapping):
            if isinstance(v, Mapping):
                r = dict_update(d.get(k, {}), v)
                d[k] = r
            else:
                d[k] = u[k]
        else:
            d = { k: u[k] }
    return d


def dict_depth(d):
    """ Get the depth of a dict
    """
    # https://stackoverflow.com/a/23499101
    if isinstance(d, dict):
        return 1 + (max(map(dict_depth, d.values())) if d else 0)
    return 0


def cf_safe_name(name : str) -> str:
    import re
    if isinstance(name, str):
        if re.match('^[0-9_]', name):
            # Add a letter to the front
            name = "v_{}".format(name)
        return re.sub(r'[^_a-zA-Z0-9]', "_", name)

    raise ValueError('Could not convert "{}" to a safe name'.format(name))


class GeoNumpyDateEncoder(geojson.GeoJSONEncoder):

    def default(self, obj : Any) -> Any:
        """If input object is an ndarray it will be converted into a list
        """
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.generic):
            return obj.item()
        # elif isinstance(obj, pd.Timestamp):
        #     return obj.to_pydatetime().isoformat()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif np.isnan(obj):
            return None

        return geojson.factory.GeoJSON.to_instance(obj)


def great_circle_distance(lat_arr, lon_arr):
    dist = np.ma.zeros(lon_arr.size, dtype=np.float64)
    g = Geod(ellps='WGS84')
    _, _, dist[1:] = g.inv(lon_arr[:-1], lat_arr[:-1], lon_arr[1:], lat_arr[1:])
    return dist



#!python
"""Download and process source data used for ConfigCreator"""
import logging
import shutil
from pathlib import Path
from urllib import request

import netCDF4 as nc
import xarray as xr
from nco import Nco

logger = logging.getLogger('ConfigCreator:get_assets.py')
logger.setLevel(logging.INFO)

SOURCES = {
    'OCEAN_ATLAS': {
        'ts_url': 'https://data.nodc.noaa.gov/thredds/fileServer/ncei/woa/{0}/A5B7/1.00/woa18_A5B7_{1}{2:02d}_01.nc',
        'other_url': 'https://data.nodc.noaa.gov/thredds/fileServer/ncei/woa/{0}/all/1.00/woa18_all_{1}{2:02d}_01.nc'
    },
    'NARR': {
        'url': 'ftp://ftp.cdc.noaa.gov/Datasets/ncep.reanalysis.derived/surface/{0}'
    }
}


def ocean_atlas_download(output_dir, month, sources=SOURCES):
    variable_map = {
        'temperature': 't',
        'salinity': 's',
        'oxygen': 'o'
    }
    for name, name_in_file in variable_map.items():
        if name in ['temperature', 'salinity']:
            url = sources['OCEAN_ATLAS']['ts_url'].format(name, name_in_file, month.month)
        else:
            url = sources['OCEAN_ATLAS']['other_url'].format(name, name_in_file, month.month)
        r = request.urlopen(url)
        data = r.read()

        fname = f'ocean_atlas_{name}_{month.month:02}.nc'
        out_file = output_dir / fname
        with open(out_file, 'wb') as f:
            f.write(data)


def ocean_atlas_merge_variables(output_dir, month):
    ocean_atlas_files = output_dir.glob(f'ocean_atlas_*_{month.month:02}.nc')
    ocean_atlas_files = list(ocean_atlas_files)
    ocean_atlas_files.sort()
    outfile = output_dir / f'ocean_atlas_{month.month:02}.nc'

    nco = Nco()
    options = [
        '-A',
        '-4',
        '-L 5'
    ]
    outfile = shutil.copy(ocean_atlas_files[0], outfile)
    for f in ocean_atlas_files[1::]:
        nco.ncks(input=str(f), output=str(outfile), options=options)


def ocean_atlas_variable_enhance(output_dir, month):
    fname = output_dir / f'ocean_atlas_{month.month:02}.nc'

    # only keep variables needed
    nco = Nco()
    vars_to_keep = ','.join(['s_an', 't_an', 'o_an'])
    options = [
        '-h',
        f'-v {vars_to_keep}'
    ]
    nco.ncks(input=str(fname), output=str(fname), options=options)

    # original time in months from 1900-01-01.
    # - force to days since 1970-01-01 for first of each month
    # - makes it easy to use with xarray
    new_time_units = 'days since 1970-01-01'
    num = nc.date2num(month, new_time_units)
    with nc.Dataset(fname, 'a') as ncds:
        tvar = ncds.variables['time']
        tvar[:] = num
        tvar.units = new_time_units
        tvar.calendar = 'standard'

    # make time a record
    options = [
        '-O',
        '--mk_rec_dmn time'
    ]
    nco.ncks(input=str(fname), output=str(fname), options=options)


def ocean_atlas_merge_time(output_dir):
    variable_merged_files = output_dir.glob('ocean_atlas_??.nc')
    variable_merged_files = [str(merged_file) for merged_file in list(variable_merged_files)]
    variable_merged_files.sort()
    output_file = output_dir.parent / 'ocean_atlas.nc'

    nco = Nco()
    options = [
        '-A'
    ]
    nco.ncrcat(input=variable_merged_files, output=str(output_file), options=options)


def ocean_atlas_enhance(output_dir):
    output_file = output_dir.parent / 'ocean_atlas.nc'
    output_tmp_file = output_dir.parent / 'ocean_atlas_tmp.nc'

    # change value that fits in a packed short var
    nco = Nco()
    options = [
        '-O',
        '-a _FillValue,,o,f,-127'
    ]
    nco.ncatted(input=str(output_file), output=str(output_tmp_file), options=options)

    # pack to use bytes
    # - requires output file defined with -o option
    options = [
        '-O',
        '-M flt_byt',
        f"-o {str(output_file)}"
    ]
    nco.ncpdq(input=str(output_tmp_file), output=str(output_file), options=options)


def get_ocean_atlas(output_dir):
    time_range = xr.cftime_range(start='2018', end='2018-12-31', freq='MS')
    for month in time_range:
        logger.info(f'downloading Ocean Atlas for {month}')
        ocean_atlas_download(output_dir, month)
        ocean_atlas_merge_variables(output_dir, month)
        ocean_atlas_variable_enhance(output_dir, month)
    ocean_atlas_merge_time(output_dir)
    ocean_atlas_enhance(output_dir)


def narr_download(output_dir, sources=SOURCES):
    variables = {
        'air': 'air.sig995.mon.ltm.nc',
        'rhum': 'rhum.sig995.mon.ltm.nc',
        'slp': 'slp.mon.ltm.nc',
        'uwind': 'uwnd.sig995.mon.ltm.nc',
        'vwind': 'vwnd.sig995.mon.ltm.nc'
    }
    for variable_name, variable_file in variables.items():
        url = sources['NARR']['url'].format(variable_file)
        r = request.urlopen(url)
        data = r.read()

        fname = f'narr_{variable_name}.nc'
        out_file = output_dir / fname
        with open(out_file, 'wb') as f:
            f.write(data)


def narr_merge_variables(output_dir):
    narr_files = output_dir.glob('narr_*.nc')
    narr_files = list(narr_files)
    narr_files.sort()
    outfile = output_dir.parent / 'narr.nc'

    nco = Nco()
    options = [
        '-A',
        '-4',
        '-L 5'
    ]
    outfile = shutil.copy(narr_files[0], outfile)
    for f in narr_files[1::]:
        nco.ncks(input=str(f), output=str(outfile), options=options)


def narr_enhance(output_dir):
    outfile = output_dir.parent / 'narr.nc'
    outtmp = output_dir.parent / 'narr_tmp.nc'

    # remove unnecessary vars
    nco = Nco()
    options = [
        '-O',
        '-C',
        '-x',
        '-v climatology_bounds,valid_yr_count'
    ]
    nco.ncks(input=str(outfile), output=str(outfile), options=options)

    # change lon from [0, 360) to [-180, 180)
    options = [
        '-O',
        '--msa',
        '-d lon,181.,360.',
        '-d lon,0.,180.0',
    ]
    nco.ncks(input=str(outfile), output=str(outtmp), options=options)

    options = [
        '-O',
        "-s 'where(lon > 180) lon=lon-360'"
    ]
    nco.ncap2(input=str(outtmp), output=str(outfile), options=options)

    # times don't work with xarray, so the year is changed to the middle of climatology
    # climatology: 1981-01-01 - 2010-12-31
    # midyear is 1996
    time_range = xr.cftime_range(start='1996', end='1996-12-31', freq='MS')
    new_units = 'days since 1970-01-01 00:00:00'
    new_times = nc.date2num(time_range, units=new_units)
    with nc.Dataset(outfile, 'a') as ncd:
        time = ncd.variables['time']
        time[:] = new_times
        time.units = new_units


def get_narr(output_dir):
    logger.info('downloading NARR')
    narr_download(output_dir)
    narr_merge_variables(output_dir)
    narr_enhance(output_dir)


def remove_tmp_files(dirs_to_delete):
    logger.info('removing tmp files')
    for dir in dirs_to_delete:
        logger.info(f'removing {dir}')
        shutil.rmtree(str(dir))


def main(output_dir, remove_tmp_files=False):
    output_dir = Path(output_dir).resolve()
    output_dir.mkdir(parents=True, exist_ok=True)

    logger.info(f'Downloading and saving data for QcConfigCreator to {output_dir}')
    logger.info('Downloading Ocean Atlas')
    ocean_atlas_dir = output_dir / 'ocean_atlas'
    ocean_atlas_dir.mkdir(exist_ok=True)
    get_ocean_atlas(ocean_atlas_dir)

    logger.info('Downloading NARR')
    narr_dir = output_dir / 'narr'
    narr_dir.mkdir(exist_ok=True)
    get_narr(narr_dir)

    if remove_tmp_files:
        remove_tmp_files([ocean_atlas_dir, narr_dir])


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser('Download source data for ConfigCreator')
    parser.add_argument(
        '-o',
        '--outdir',
        type=str,
        default='resources',
        help='Output path for downloaded data'
    )
    parser.add_argument(
        '-d',
        '--delete',
        action='store_true',
        help='Remove temporary files'
    )
    args = parser.parse_args()
    main(args.outdir, args.delete)



#!/usr/bin/env python
# coding=utf-8
import inspect
import logging
from typing import List
import simplejson as json
from pathlib import Path
from importlib import import_module

import numpy as np
import pandas as pd
import netCDF4 as nc4

from ioos_qc.config import Config
from ioos_qc.qartod import aggregate
from ioos_qc.utils import GeoNumpyDateEncoder, cf_safe_name
from ioos_qc.results import collect_results, CollectedResult


L = logging.getLogger(__name__)  # noqa


class BaseStore:

    def save(self, *args, **kwargs):
        """
        Serialize results to a store. This could save a file or publish messages.
        """
        pass

    @property
    def stream_ids(self) -> List[str]:
        """
        A list of stream_ids to save to the store
        """
        pass


class PandasStore(BaseStore):
    """Store results in a dataframe"""

    def __init__(self, results, axes: dict = None):
        # OK, time to evaluate the actual tests now that we need the results
        self.results = list(results)
        self.collected_results = collect_results(self.results, how='list')
        self._stream_ids = [ cr.stream_id for cr in self.collected_results ]
        self.axes = axes or {
            't': 'time',
            'z': 'z',
            'y': 'lat',
            'x': 'lon'
        }

    @property
    def stream_ids(self) -> List[str]:
        return self._stream_ids

    def compute_aggregate(self, name='rollup'):
        """ Internally compute the total aggregate and add it to the results
        """
        agg = CollectedResult(
            stream_id='',
            package='qartod',
            test=name,
            function=aggregate,
            results=aggregate(self.collected_results)
        )
        self.collected_results.append(agg)

    def save(self,
             write_data: bool = False,
             write_axes: bool = True,
             include: list = None,
             exclude: list = None) -> pd.DataFrame:

        df = pd.DataFrame()

        for cr in self.collected_results:

            # Add time axis
            if write_axes is True and self.axes['t'] not in df and cr.tinp is not None and cr.tinp.size != 0:
                L.info(f"Adding column {self.axes['t']} from stream {cr.stream_id}")
                df[self.axes['t']] = cr.tinp

            # Add z axis
            if write_axes is True and self.axes['z'] not in df and cr.zinp is not None and cr.zinp.size != 0:
                L.info(f"Adding column {self.axes['z']} from stream {cr.stream_id}")
                df[self.axes['z']] = cr.zinp

            # Add x axis
            if write_axes is True and self.axes['x'] not in df and cr.lon is not None and cr.lon.size != 0:
                L.info(f"Adding column {self.axes['x']} from stream {cr.stream_id}")
                df[self.axes['x']] = cr.lon

            # Add x axis
            if write_axes is True and self.axes['y'] not in df and cr.lat is not None and cr.lat.size != 0:
                L.info(f"Adding column {self.axes['y']} from stream {cr.stream_id}")
                df[self.axes['y']] = cr.lat

            # Inclusion list, skip everything not defined
            if include is not None and (cr.function not in include and cr.stream_id not in include and cr.test not in include):
                continue

            # Exclusion list, skip everything defined
            if exclude is not None and (cr.function in exclude or cr.stream_id in exclude or cr.test in cr.test in include):
                continue

            # Add data column
            if write_data and cr.stream_id not in df:
                df[cr.stream_id] = cr.data

            # Add QC results column
            # Aggregate will have None stream_id, so allow it to be that way!
            stream_label = f'{cr.stream_id}.' if cr.stream_id else ''
            package_label = f'{cr.package}.' if cr.package else ''
            test_label = f'{cr.test}' if cr.test else ''
            column_name = cf_safe_name(f'{stream_label}{package_label}{test_label}')
            if column_name not in df:
                df[column_name] = cr.results
            else:
                L.warning(f"Found duplicate QC results column: {column_name}, skipping.")

        return df


class CFNetCDFStore(BaseStore):

    def __init__(self, results, axes=None, **kwargs):
        # OK, time to evaluate the actual tests now that we need the results
        self.results = list(results)
        self.collected_results = collect_results(self.results, how='list')
        self._stream_ids = [ cr.stream_id for cr in self.collected_results ]
        self.axes = axes or {
            't': 'time',
            'z': 'z',
            'y': 'lat',
            'x': 'lon'
        }

    @property
    def stream_ids(self) -> List[str]:
        return self._stream_ids

    def save(self, path_or_ncd, dsg, config: Config, dsg_kwargs: dict = {}, write_data: bool = False, include: list = None, exclude: list = None):
        ps = PandasStore(self.results, self.axes)
        df = ps.save(write_data=write_data, include=include, exclude=exclude)

        # Write a new file
        attrs = {}
        for cr in self.collected_results:

            column_name = cf_safe_name(f'{cr.stream_id}.{cr.package}.{cr.test}')

            # Set the ancillary variables
            if cr.stream_id not in attrs:
                attrs[cr.stream_id] = {
                    'ancillary_variables': column_name
                }
            else:
                # Update the source ancillary_variables
                existing = getattr(attrs[cr.stream_id], 'ancillary_variables', '').split(' ')
                existing += [column_name]
                attrs[cr.stream_id] = ' '.join(list(set(existing))).strip()

            # determine standard name and long name. These should be defined on each test function
            # https://github.com/cf-convention/cf-conventions/issues/216
            standard_name = getattr(cr.function, 'standard_name', 'quality_flag')
            long_name = getattr(cr.function, 'long_name', 'Quality Flag')

            # Get flags from module attribute called FLAGS
            flags = getattr(inspect.getmodule(cr.function), 'FLAGS')
            varflagnames = [ d for d in flags.__dict__ if not d.startswith('__') ]
            varflagvalues = [ getattr(flags, d) for d in varflagnames ]

            # Set QC variable attributes
            if column_name not in attrs:
                attrs[column_name] = {
                    'standard_name': standard_name,
                    'long_name': long_name,
                    'flag_values': np.byte(varflagvalues),
                    'flag_meanings': ' '.join(varflagnames),
                    'valid_min': np.byte(min(varflagvalues)),
                    'valid_max': np.byte(max(varflagvalues)),
                    'ioos_qc_module': cr.package,
                    'ioos_qc_test': cr.test,
                    'ioos_qc_target': cr.stream_id,
                }
                # If there is only one context we can write variable specific configs
                # We can't do this across different contexts and this would repeat the regions
                # and windows for each variable even if they are equal. This needs another look.
                if len(config.contexts) == 1:
                    calls = config.calls_by_stream_id(column_name)

                    if not calls:
                        # No stream_id found!
                        continue

                    # Use the first call of this stream_id. There will be only 1 because there
                    # is only one context
                    call = calls[0]
                    if call.region:
                        attrs[column_name]['ioos_qc_region'] = json.dumps(
                            call.region,
                            cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True
                        )
                    if call.window.starting or call.window.ending:
                        attrs[column_name]['ioos_qc_window'] = json.dumps(
                            call.window,
                            cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True
                        )

                    qc_varconfig = json.dumps(
                        call.config(),
                        cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True
                    )
                    attrs[column_name]['ioos_qc_config'] = qc_varconfig

        if len(config.contexts) > 1:
            # We represent the config as one global config JSON object
            attrs['ioos_qc_config'] = json.dumps(
                config.config, cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True
            )

        dsg_kwargs = {
            **dsg_kwargs,
            **{
                'attributes': attrs
            }
        }

        # pocean requires these default columns, which should be removed as a requirement
        # in pocean.
        df['station'] = 0
        df['trajectory'] = 0
        df['profile'] = 0
        if 'z' not in df:
            df['z'] = 0
        ncd = dsg.from_dataframe(df, path_or_ncd, axes=self.axes, **dsg_kwargs)
        return ncd


class NetcdfStore:

    def save(self, path_or_ncd, config, results):
        """
        Updates the given netcdf with test configuration and results.
        If there is already a variable for a given test, it will update that variable with the latest results.
        Otherwise, it will create a new variable.

        :param path_or_ncd: path or netcdf4 Dataset in which to store results
        :param results: output of run()
        """
        try:
            ncd = None
            should_close = True
            if isinstance(path_or_ncd, (str, Path)):
                ncd = nc4.Dataset(str(path_or_ncd), 'a')
            elif isinstance(path_or_ncd, nc4.Dataset):
                ncd = path_or_ncd
                should_close = False
            else:
                return ValueError('Input is not a valid file path or Dataset')

            for vname, qcobj in results.items():

                if vname not in ncd.variables:
                    L.warning(f'{vname} not found in the Dataset, skipping')
                    continue

                source_var = ncd.variables[vname]
                # Keep track of the test names so we can add to the source's
                # ancillary_variables at the end
                qcvar_names = []

                for modu, tests in qcobj.items():

                    try:
                        testpackage = import_module('ioos_qc.{}'.format(modu))
                    except ImportError:
                        L.error('No ioos_qc test package "{}" was found, skipping.'.format(modu))
                        continue

                    for testname, testresults in tests.items():

                        # Try to find a qc variable that matches this config
                        qcvars = ncd.get_variables_by_attributes(
                            ioos_qc_module=modu,
                            ioos_qc_test=testname,
                            ioos_qc_target=vname
                        )
                        if not qcvars:
                            qcvarname = cf_safe_name(vname + '.' + modu + '.' + testname)
                        else:
                            if len(qcvars) > 1:
                                names = [ v.name for v in qcvars ]
                                L.warning('Found more than one QC variable match: {}'.format(names))
                            # Use the last one found
                            qcvarname = qcvars[-1].name

                        # Get flags from module attribute called FLAGS
                        flags = getattr(testpackage, 'FLAGS')
                        varflagnames = [ d for d in flags.__dict__ if not d.startswith('__') ]
                        varflagvalues = [ getattr(flags, d) for d in varflagnames ]

                        if qcvarname not in ncd.variables:
                            v = ncd.createVariable(qcvarname, np.byte, source_var.dimensions)
                        else:
                            v = ncd[qcvarname]
                        qcvar_names.append(qcvarname)

                        # determine standard name
                        # https://github.com/cf-convention/cf-conventions/issues/216
                        try:
                            testfn = getattr(testpackage, testname)
                            standard_name = testfn.standard_name
                            long_name = testfn.long_name
                        except AttributeError:
                            standard_name = 'quality_flag'
                            long_name = 'Quality Flag'

                        # write to netcdf
                        v[:] = testresults
                        v.setncattr('standard_name', standard_name)
                        v.setncattr('long_name', long_name)
                        v.setncattr('flag_values', np.byte(varflagvalues))
                        v.setncattr('flag_meanings', ' '.join(varflagnames))
                        v.setncattr('valid_min', np.byte(min(varflagvalues)))
                        v.setncattr('valid_max', np.byte(max(varflagvalues)))
                        v.setncattr('ioos_qc_module', modu)
                        v.setncattr('ioos_qc_test', testname)
                        v.setncattr('ioos_qc_target', vname)
                        # If there is only one context we can write variable specific configs
                        if len(config.contexts) == 1:
                            varconfig = config.contexts[0].streams[vname].config[modu][testname]
                            varconfig = json.dumps(varconfig, cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True)
                            v.setncattr('ioos_qc_config', varconfig)
                            v.setncattr('ioos_qc_region', json.dumps(config.contexts[0].region, cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True))
                            v.setncattr('ioos_qc_window', json.dumps(config.contexts[0].window, cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True))

                # Update the source ancillary_variables
                existing = getattr(source_var, 'ancillary_variables', '').split(' ')
                if qcvar_names:
                    existing += qcvar_names
                source_var.ancillary_variables = ' '.join(list(set(existing))).strip()

            if len(config.contexts) > 1:
                # We can't represent these at the variable level, so make one global config
                ncd.setncattr(
                    'ioos_qc_config',
                    json.dumps(config.config, cls=GeoNumpyDateEncoder, allow_nan=False, ignore_nan=True)
                )

        finally:
            if ncd and should_close is True:
                ncd.close()



#!/usr/bin/env python
# coding=utf-8
"""Tests based on the IOOS QARTOD manuals."""
import logging
from typing import Tuple, Sequence
from collections import namedtuple

import numpy as np

from ioos_qc.utils import (
    isnan,
    add_flag_metadata,
    mapdates
)
from ioos_qc.qartod import QartodFlags

L = logging.getLogger(__name__)  # noqa

FLAGS = QartodFlags  # Default name for all check modules
NOTEVAL_VALUE = QartodFlags.UNKNOWN

span = namedtuple('Span', 'minv maxv')


@add_flag_metadata(standard_name='gross_range_test_quality_flag',
                   long_name='Gross Range Test Quality Flag')
def valid_range_test(inp : Sequence[any],
                     valid_span : Tuple[any, any],
                     dtype : np.dtype = None,
                     start_inclusive : bool = True,
                     end_inclusive : bool = False,
                     ) -> np.ma.core.MaskedArray:
    """Checks that values are within a min/max range. This is not unlike a `qartod.gross_range_test`
    with fail and suspect bounds being equal, except that here we specify the inclusive range that
    should pass instead of the exclusive bounds which should fail. This also supports datetime-like
    objects where as the `qartod.gross_range_test` method only supports numerics.

    Given a 2-tuple of minimum/maximum values, flag data outside of the given
    range as FAIL data. Missing and masked data is flagged as UNKNOWN. The first span value is
    treated as inclusive and the second span valid is treated as exclusive. To change this
    behavior you can use the parameters `start_inclusive` and `end_inclusive`.

    Args:
        inp (Sequence[any]): Data as a sequence of objects compatible with the fail_span objects
        fail_span (Tuple[any, any]): 2-tuple range which to flag outside data as FAIL. Objects
            should be of equal format to that of the inp parameter as they will be checked for
            equality without type conversion.
        dtype (np.dtype): Optional. If your data is not already numpy-typed you can specify its
            dtype here.
        start_inclusive (bool): Optional. If the starting span value should be inclusive (True) or
            exclusive (False).
        end_inclusive (bool): Optional. If the ending span value should be inclusive (True) or
            exclusive (False).

    Returns:
        np.ma.core.MaskedArray: A masked array of flag values equal in size to that of the input.
    """

    # Numpy array inputs
    if dtype is None and hasattr(inp, 'dtype'):
        dtype = inp.dtype
    # Pandas column inputs
    # This is required because we don't want to restrict a user from using a pd.Series
    # directly with this function. If the data was coming from a Store, it would
    # always be a numpy array.
    elif dtype is None and hasattr(inp, 'values') and hasattr(inp.values, 'dtype'):
        dtype = inp.values.dtype

    # Save original shape
    original_shape = inp.shape

    if dtype is None:
        L.warning("Trying to guess data input type, try specifying the dtype parameter")
        # Try to figure out the dtype so masked values can be calculated
        try:
            # Try datetime-like objects
            inp = np.ma.masked_invalid(mapdates(inp))
            valid_span = np.ma.masked_invalid(mapdates(valid_span))
        except BaseException:
            try:
                # Try floating point numbers
                inp = np.ma.masked_invalid(np.array(inp).astype(np.floating))
                valid_span = np.ma.masked_invalid(np.array(valid_span).astype(np.floating))
            except BaseException:
                # Well, we tried.
                raise ValueError(
                    "Could not determine the type of input, try using the dtype parameter")
    else:
        inp = np.ma.masked_invalid(np.array(inp, dtype=dtype))
        valid_span = np.ma.masked_invalid(np.array(valid_span, dtype=dtype))

    inp = inp.flatten()

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype='uint8')

    # Set fail on either side of the bounds, inclusive and exclusive
    if not isnan(valid_span[0]):
        with np.errstate(invalid='ignore'):
            if start_inclusive is True:
                flag_arr[inp < valid_span[0]] = QartodFlags.FAIL
            else:
                flag_arr[inp <= valid_span[0]] = QartodFlags.FAIL

    if not isnan(valid_span[1]):
        with np.errstate(invalid='ignore'):
            if end_inclusive is True:
                flag_arr[inp > valid_span[1]] = QartodFlags.FAIL
            else:
                flag_arr[inp >= valid_span[1]] = QartodFlags.FAIL

    # If the value is masked or nan set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)



#!/usr/bin/env python
# coding=utf-8
import logging
from typing import NamedTuple, List
from dataclasses import dataclass
from collections import OrderedDict as odict, defaultdict

import numpy as np
from ioos_qc.qartod import QartodFlags

L = logging.getLogger(__name__)  # noqa


class CallResult(NamedTuple):
    package: str
    test: str
    function: callable
    results: np.ndarray

    def __repr__(self):
        return f'<CallResult package={self.package} test={self.test}>'


class ContextResult(NamedTuple):
    stream_id: str
    results: List[CallResult]
    subset_indexes: np.ndarray
    data: np.ndarray = None
    tinp: np.ndarray = None
    zinp: np.ndarray = None
    lat: np.ndarray = None
    lon: np.ndarray = None

    def __repr__(self):
        return f'<ContextResult stream_id={self.stream_id}>'


@dataclass
class CollectedResult:
    stream_id: str
    package: str
    test: str
    function: callable
    results: np.ma.core.MaskedArray = None
    data: np.ndarray = None
    tinp: np.ndarray = None
    zinp: np.ndarray = None
    lat: np.ndarray = None
    lon: np.ndarray = None

    def __repr__(self):
        return f'<CollectedResult stream_id={self.stream_id} package={self.package} test={self.test}>'

    def function_name(self) -> str:
        return self.function.__name__

    @property
    def hash_key(self) -> str:
        return f'{self.stream_id}:{self.package}.{self.test}'


def collect_results(results, how='list'):
    if how in ['list', list]:
        return collect_results_list(results)
    elif how in ['dict', dict]:
        return collect_results_dict(results)


def collect_results_list(results):
    """ Turns a list of ContextResult objects into an iterator of CollectedResult objects
        by combining the subset_index information in each ContextResult together into
        a single array of results.
    """
    collected = odict()

    # ContextResults
    for r in results:

        cr = None
        # Shortcut for CallResult objects when someone uses QcConfig.run() directly
        # and doesn't go through a Stream object
        if isinstance(r, CallResult):
            cr = CollectedResult(
                stream_id=None,
                package=r.package,
                test=r.test,
                function=r.function,
                results=r.results,
            )
            collected[cr.hash_key] = cr
            continue

        # CallResults
        for tr in r.results:

            cr = CollectedResult(
                stream_id=r.stream_id,
                package=tr.package,
                test=tr.test,
                function=tr.function
            )

            if cr.hash_key not in collected:
                # Set the initial values
                cr.results = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=tr.results.dtype)
                cr.data = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=r.data.dtype)
                cr.tinp = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=r.tinp.dtype)
                cr.zinp = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=r.zinp.dtype)
                cr.lat = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=r.lat.dtype)
                cr.lon = np.ma.masked_all(shape=r.subset_indexes.shape, dtype=r.lon.dtype)
                collected[cr.hash_key] = cr

            collected[cr.hash_key].results[r.subset_indexes] = tr.results

        if cr is not None:
            if r.subset_indexes.all():
                collected[cr.hash_key].data = r.data
                collected[cr.hash_key].tinp = r.tinp
                collected[cr.hash_key].zinp = r.zinp
                collected[cr.hash_key].lat = r.lat
                collected[cr.hash_key].lon = r.lon
            else:
                collected[cr.hash_key].data[r.subset_indexes] = r.data
                collected[cr.hash_key].tinp[r.subset_indexes] = r.tinp
                collected[cr.hash_key].zinp[r.subset_indexes] = r.zinp
                collected[cr.hash_key].lat[r.subset_indexes] = r.lat
                collected[cr.hash_key].lon[r.subset_indexes] = r.lon

    return list(collected.values())


def collect_results_dict(results):
    """ Turns a list of ContextResult objects into a dictionary of test results
        by combining the subset_index information in each ContextResult together into
        a single array of results. This is mostly here for historical purposes. Users
        should migrate to using the Result objects directly.
    """
    # Magic for nested key generation
    # https://stackoverflow.com/a/27809959
    collected = defaultdict(lambda: defaultdict(odict))

    # ContextResults
    for r in results:

        # Shortcut for CallResult objects when someone uses QcConfig.run() directly
        # and doesn't go through a Stream object
        if isinstance(r, CallResult):
            collected[r.package][r.test] = r.results
            continue

        flag_arr = np.ma.empty_like(r.subset_indexes, dtype='uint8')
        flag_arr.fill(QartodFlags.UNKNOWN)

        # iterate over the CallResults
        for tr in r.results:
            testpackage = tr.package
            testname = tr.test
            testresults = tr.results

            if testname not in collected[r.stream_id][testpackage]:
                collected[r.stream_id][testpackage][testname] = np.copy(flag_arr)
            collected[r.stream_id][testpackage][testname][r.subset_indexes] = testresults

    return collected



import calendar
import datetime
import json
import logging
from copy import copy
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from jsonschema import validate
from scipy.interpolate import CubicSpline

from ioos_qc.config_creator import fx_parser

L = logging.getLogger(__name__)  # noqa


QC_CONFIG_CREATOR_SCHEMA = {
    "type": "object",
    "title": "QcConfigCreator Schema",
    "description": "Schema to validate configuration for QcCreatorConfig",
    "required": [
        "datasets"
    ],
    "definitions": {
        "variable": {
            "type": "object",
            "title": "Variable map (variable name in QcConfig -> variable name in dataset",
            "items": {"type": "object"}
        },
        "dataset": {
            "type": "object",
            "title": "Dataset description",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of dataset"
                },
                "file_path": {
                    "type": "string",
                    "description": "Path to file used to create configuration."
                },
                "variables": {
                    "type": "object",
                    "description": "Variables in file used to create configuration."
                },
                "3d": {
                    "type": "string",
                    "description": "Include if 3d dataset with value being the name of the 3rd dimension"
                }
            },
            "required": [
                "name",
                "file_path",
                "variables"
            ],
        },
    },
    "properties": {
        "datasets": {
            "type": "array",
            "description": "Array of datasets used for QcCreatorConfig",
            "items": {
                "$ref": "#/definitions/dataset"
            }
        }
    }
}

VARIABLE_CONFIG_SCHEMA = {
    "type": "object",
    "title": "QcVariableConfig",
    "description": "Schema to validate configuration for QcVariableConfig",
    "required": [
        "variable",
        "bbox",
        "start_time",
        "end_time",
        "tests"
    ],
    "definitions": {
        "test": {
            "type": "object",
            "description": "Test (gross_range_test, etc.) definition",
            "required": [
                "suspect_min",
                "suspect_max",
                "fail_min",
                "fail_max"
            ],
            "properties": {
                "suspect_min": {
                    "type": "string",
                    "description": "Function or value to define suspect min"
                },
                "suspect_max": {
                    "type": "string",
                    "description": "Function or value to define suspect max"
                },
                "fail_min": {
                    "type": "string",
                    "description": "Function or value to define fail min"
                },
                "fail_max": {
                    "type": "string",
                    "description": "Function or value to define fail max"
                }
            }
        }
    },
    "properties": {
        "variable": {
            "type": "string",
            "description": "Variable name. The same name used as a key in 'variable' object in CreatorConfig"
        },
        "bbox": {
            "type": "array",
            "description": "Bounding box of region in deployment (xmin, ymin, xmax, ymax)"
        },
        "start_time": {
            "type": "string",
            "description": "Start time of deployment (YYYY-MM-DD)"
        },
        "end_time": {
            "type": "string",
            "description": "Exclusive end time of deployment (YYYY-MM-DD)"
        },
        "tests": {
            "type": "object",
            "items": {
                "$ref": "#/definitions/test",
            }
        }
    }
}


class CreatorConfig(dict):
    """
    Defines the dataset(s) configuration used by QcConfigCreator.

    :param path_or_dict: QcConfigCreator configuration, one of the following formats:
        python dict
        JSON filepath (str or Path object)
    :param dict: JSON schema for CreatorConfig
    """
    def __init__(self, path_or_dict, schema=QC_CONFIG_CREATOR_SCHEMA):
        if isinstance(path_or_dict, str) or isinstance(path_or_dict, Path):
            with open(path_or_dict) as f:
                config = json.load(f)
        elif isinstance(path_or_dict, dict):
            config = path_or_dict
        else:
            raise ValueError('Input is not valid file path or dict')
        validate(instance=config, schema=schema)

        datasets = {}
        for dataset in config['datasets']:
            if '3d' in dataset:
                datasets[dataset['name']] = {
                    'file_path': dataset['file_path'],
                    'variables': dataset['variables'],
                    '3d': dataset['3d']
                }
            else:
                datasets[dataset['name']] = {
                    'file_path': dataset['file_path'],
                    'variables': dataset['variables'],
                }
        self.update(datasets)

    def __str__(self):
        return json.dumps(self, indent=4, sort_keys=True)


class QcVariableConfig(dict):
    """
    Used to generate a QcConfig for a specific variable.

    Args:
        path_or_dict: QcVariableConfig configuration, one of the following formats:
            python dict
            JSON filepath (str or Path object)
        schema: JSON schema for QcVariable
    """
    allowed_stats = [
        'min',
        'max',
        'mean',
        'std'
    ]
    allowed_operators = [
        '+',
        '-',
        '*',
        '/',
    ]
    allowed_groupings = [
        '(',
        ')'
    ]

    def __init__(self, path_or_dict, schema=VARIABLE_CONFIG_SCHEMA):
        if isinstance(path_or_dict, str) or isinstance(path_or_dict, Path):
            with open(path_or_dict) as f:
                config = json.load(f)
        elif isinstance(path_or_dict, dict):
            config = path_or_dict
        else:
            raise ValueError('Input is not valid file path or dict')

        L.debug("Validating schema...")
        validate(instance=config, schema=schema)

        # validate test specifications only contain allowed stats and operators
        for test in config['tests'].keys():
            for test_name, test_def in config['tests'][test].items():
                if test_name == 'bbox':
                    continue
                self._validate_fx(test_def, test_name)

        self.update(config)

    def _validate_fx(self, input_fx, test_name):
        """Thows exception if input_fx contains tokens not specifically allowed"""
        tokens = input_fx.split(' ')
        for token in tokens:
            try:
                _ = float(token)
            except ValueError:
                if token not in self.allowed_stats and \
                   token not in self.allowed_operators and \
                   token not in self.allowed_groupings:
                    msg = (
                        f'{token} not allowed in min/max specification in config of {test_name}.\n'
                        f'Allowable stats are: {[s for s in self.allowed_stats]}.\n'
                        f'Allowable operators are: {[o for o in self.allowed_operators]}.'
                        f'Allowable groupings are: {[o for o in self.allowed_groupings]}.'
                    )
                    raise ValueError(msg)

    def __str__(self):
        return json.dumps(self, indent=4, sort_keys=True)


class QcConfigCreator:
    """Creates level-0 configuration to create QcQonfig.

    Arguments:
        creator_config (QcCreatorConfig): Configuration for datasets and variables used to create qc_config.

    Attributes:
        allowed_stats (list): Specific statistics allowed to be used to configure each test.
        allowed_operators (list): Operators allowed to used to configure each test.
    """

    def __init__(self, creator_config):
        self.config = creator_config
        self.datasets = self._load_datasets()
        self.dataset_years = self._determine_dataset_years()

    def create_config(self, variable_config):
        """Create QARTOD QC config given QcVariableConfig.

        Args:
            variable_config (QcVariableConfig): Config for variable to be quality controlled

        Returns:
            qc_config (dict): Config for ioos_qc
        """
        stats = self._get_stats(variable_config)
        L.debug("Creating config...")
        test_configs = {
            name: self._create_test_section(name, variable_config, stats) for name in variable_config['tests'].keys()
        }

        return {
            f'{variable_config["variable"]}': {
                'qartod': test_configs
            }
        }

    def _load_datasets(self):
        """Load datasets"""
        L.debug(f"Loading {len(self.config)} datasets...")
        return {name: xr.load_dataset(self.config[name]['file_path']) for name in self.config.keys()}

    def _determine_dataset_years(self):
        """Determine year used in datasets, return as dict {dataset_name, year}.

        Notes:
            - Each dataset is from a unique climatology or source,
              so the monthly files have different years.
        """
        years = {}
        for dataset_name, dataset in self.datasets.items():
            years[dataset_name] = dataset['time.year'][0].data

        return years

    def _var2var_in_file(self, var):
        """Return variable name used in the dataset and dataset name"""
        for dataset_name, dataset in self.config.items():
            if var in dataset['variables'].keys():
                return dataset['variables'][var], dataset_name

    def var2dataset(self, var):
        """Return dataset name and dataset for given variable (as named in qc_config, not in the file)"""
        _, dataset_name = self._var2var_in_file(var)

        return dataset_name, self.datasets[dataset_name]

    def _create_test_section(self, test_name, variable_config, test_limits):
        """Given test_name, QcVariableConfig and test_limits, return qc_config section for that test."""
        if test_name == 'spike_test':
            return self.__create_spike_section(test_name, variable_config, test_limits)
        elif test_name == 'location_test':
            return self.__create_location_section(test_name, variable_config)
        elif test_name == 'rate_of_change_test':
            return self.__create_rate_of_change_section(test_name, variable_config, test_limits)
        elif test_name == 'flat_line_test':
            return self.__create_flat_line_section(test_name, variable_config, test_limits)
        else:
            return self.__create_span_section(test_name, variable_config, test_limits)

    def __create_span_section(self, test_name, variable_config, stats):
        suspect_min = fx_parser.eval_fx(variable_config['tests'][test_name]['suspect_min'], stats)
        suspect_max = fx_parser.eval_fx(variable_config['tests'][test_name]['suspect_max'], stats)
        fail_min = fx_parser.eval_fx(variable_config['tests'][test_name]['fail_min'], stats)
        fail_max = fx_parser.eval_fx(variable_config['tests'][test_name]['fail_max'], stats)

        return {
            'suspect_span': [suspect_min, suspect_max],
            'fail_span': [fail_min, fail_max]
        }

    def __create_spike_section(self, test_name, variable_config, stats):
        suspect_threshold = fx_parser.eval_fx(variable_config['tests'][test_name]['suspect_threshold'], stats)
        fail_threshold = fx_parser.eval_fx(variable_config['tests'][test_name]['fail_threshold'], stats)

        return {
            'suspect_threshold': suspect_threshold,
            'fail_threshold': fail_threshold
        }

    def __create_flat_line_section(self, test_name, variable_config, stats):
        suspect_threshold = fx_parser.eval_fx(variable_config['tests'][test_name]['suspect_threshold'], stats)
        fail_threshold = fx_parser.eval_fx(variable_config['tests'][test_name]['fail_threshold'], stats)
        tolerance = fx_parser.eval_fx(variable_config['tests'][test_name]['tolerance'], stats)

        return {
            'suspect_threshold': suspect_threshold,
            'fail_threshold': fail_threshold,
            'tolerance': tolerance
        }

    def __create_location_section(self, test_name, variable_config):
        return {
            'bbox': variable_config['tests'][test_name]['bbox']
        }

    def __create_rate_of_change_section(self, test_name, variable_config, stats):
        threshold = fx_parser.eval_fx(variable_config['tests'][test_name]['threshold'], stats)
        return {
            'threshold': threshold
        }

    def _get_stats(self, variable_config):
        """Return dict of stats (min, max, mean, std) for given config"""
        start_time = datetime.datetime.strptime(variable_config['start_time'], '%Y-%m-%d')
        end_time = datetime.datetime.strptime(variable_config['end_time'], '%Y-%m-%d')
        time_range = slice(start_time, end_time)
        subset = self._get_subset(
            variable_config['variable'],
            variable_config['bbox'],
            time_range
        )

        return {
            'min': np.nanmin(subset),
            'max': np.nanmax(subset),
            'mean': np.nanmean(subset),
            'std': np.nanstd(subset)
        }

    def _get_subset(self, var, bbox, time_slice, depth=0, pad_delta=0.5):
        """Get subset of data"""
        _, ds = self.var2dataset(var)

        lat_mask = np.logical_and(
            ds['lat'] >= bbox[1],
            ds['lat'] <= bbox[3]
        )
        lon_mask = np.logical_and(
            ds['lon'] >= bbox[0],
            ds['lon'] <= bbox[2]
        )

        # if there is no data in the subset, increase bounding box in an iterative fashion
        # - both are interpolated to daily values
        L.debug(f"Subsetting {var} by depth={depth} and {bbox}...")
        subset = self.__get_daily_interp_subset(var, time_slice, depth, lat_mask, lon_mask)

        padded = 0
        while np.nansum(subset) == 0:
            old_bbox = copy(bbox)
            bbox = self.__apply_bbox_pad(bbox, pad_delta)

            if old_bbox == bbox:
                L.warning(f"No data found for {var} at depth={depth}")
                break

            lat_mask = np.logical_and(
                ds['lat'] >= bbox[1],
                ds['lat'] <= bbox[3]
            )
            lon_mask = np.logical_and(
                ds['lon'] >= bbox[0],
                ds['lon'] <= bbox[2]
            )

            padded += 1
            subset = self.__get_daily_interp_subset(var, time_slice, depth, lat_mask, lon_mask)

        L.info(f'Used bounding box: {bbox}')
        return subset

    def __apply_bbox_pad(self, bbox, pad):
        # Prevent infinite attempts of expanding bounding box for valid data
        if bbox == ['-180', '-90', '180', '90']:
            raise RuntimeError(f'No valid data found in maximum bounding box {bbox}')

        def apply_pad(val, lat_or_lon, min_or_max):
            if lat_or_lon == 'lat':
                if min_or_max == 'min':
                    return max(-90, val - pad)
                else:
                    return min(90, val + pad)
            else:
                if val <= 0 and min_or_max == 'min':
                    return max(-180, val - pad)
                elif val <= 0 and min_or_max == 'max':
                    return min(180, val + pad)
                elif val > 0 and min_or_max == 'min':
                    return max(-180, val - pad)
                else:
                    return min(180, val + pad)

        new_bbox = []
        new_bbox.append(apply_pad(bbox[0], 'lon', 'min'))
        new_bbox.append(apply_pad(bbox[1], 'lat', 'min'))
        new_bbox.append(apply_pad(bbox[2], 'lon', 'max'))
        new_bbox.append(apply_pad(bbox[3], 'lat', 'max'))

        L.debug(f'Expanded bounding box to {bbox}')

        return new_bbox

    def __get_daily_interp_subset(self, var, time_slice, depth, lat_mask, lon_mask):
        ds_name, ds = self.var2dataset(var)
        var_in_file, _ = self._var2var_in_file(var)

        if '3d' in self.config[ds_name]:
            var = ds[var_in_file][:, depth, lat_mask, lon_mask]
        else:  # 2d
            var = ds[var_in_file][:, lat_mask, lon_mask]

        # try cubic interpolation to daily values
        try:
            return self.__daily_cubic_interp(var, time_slice)
        except ValueError:  # Raises exception is subset in bbox is all NaNs
            return 0

    def __daily_cubic_interp(self, var, time_slice):
        """Given DataArray and time_slice, interpolate to daily values and return NumPy array.

        Notes:
        - Uses day of year for interpolation
        - Adds data for day 1 and 366 to ensure periodicity if required to ensure periodicity
        """
        if (time_slice.stop - time_slice.start).days > 365:
            raise NotImplementedError('Maximum of 365 days available for config_creator')

        x = var.time.dt.dayofyear
        y = var.data
        if 1 not in x and 366 not in x:
            x = np.hstack([1, x, 366])
            # add new average value at beginning and end to ensure periodicity
            end_value = (y[0:1, :, :] + y[-1:, :, :]) / 2
            y = np.concatenate([end_value, y, end_value])
        elif 1 not in x:
            x = np.hstack([1, x])
            # add new beginning value that is equal to end value to ensure periodicity
            y = np.concatenate([y[-1:, :, :], y])
        else:  # 366 must not be in x, but 1 is
            x = np.hstack([x, 366])
            # add new end value that is equal to beginning value to ensure periodicity
            y = np.concatenate([y, y[0:1, :, :]])

        # Create interpolator
        # Need to handle NaNs in data before applying interpolator
        # - mask out NaNs
        # - reshape masked y to (ntimes, ...) where ntimes is y.shape[0]
        # - assume that NaNs are same shape through time (i.e. will not work with wetting-drying)
        y_no_nans = y[~np.isnan(y)].reshape(y.shape[0], -1)
        spline = CubicSpline(x, y_no_nans, bc_type='periodic')

        # Get days of year for independent variable
        # - special handling for when time-slice crosses new year
        doy_start = pd.Timestamp(time_slice.start).dayofyear
        doy_end = pd.Timestamp(time_slice.stop).dayofyear
        if doy_start > doy_end:  # days cross new year
            days = list(range(doy_start, 367)) + list(range(1, doy_end))
            if not calendar.isleap(time_slice.start.year):
                days.remove(366)
        else:
            days = list(range(doy_start, doy_end))

        return spline(days)

    def __str__(self):
        return json.dumps(self.config, indent=4, sort_keys=True)

    def __rpr__(self):
        return self.__str__


def to_json(qc_config, out_file=None):
    """Given qc_config return json"""
    if out_file:
        with open(out_file, 'w') as outfile:
            json.dump(outfile, qc_config)
    else:
        return json.dumps(qc_config)



import logging
import numpy as np

from bokeh import plotting
from bokeh.layouts import gridplot

L = logging.getLogger(__name__)


def bokeh_plot(data, var_name, results, title, module, test_name):
    plot = bokeh_plot_var(data, var_name, results, title, module, test_name)
    return gridplot([[plot]], sizing_mode='fixed')


def bokeh_plot_var(time, data, var_name, results, title, module, test_name):
    """ Method to plot QC results using Bokeh """

    if module not in results or test_name not in results[module]:
        L.warning(f'No results for test {module}.{test_name} found')
        return

    qc_test = results[module][test_name]

    qc_pass = np.ma.masked_where(qc_test != 1, data)
    qc_suspect = np.ma.masked_where(qc_test != 3, data)
    qc_fail = np.ma.masked_where(qc_test != 4, data)
    qc_notrun = np.ma.masked_where(qc_test != 2, data)

    p1 = plotting.figure(x_axis_type="datetime", title=test_name + ' : ' + title)
    p1.grid.grid_line_alpha = 0.3
    p1.xaxis.axis_label = 'Time'
    p1.yaxis.axis_label = 'Data'

    p1.line(time, data,  legend_label='data', color='#A6CEE3')
    p1.circle(time, qc_notrun, size=2, legend_label='qc not run', color='gray', alpha=0.2)
    p1.circle(time, qc_pass, size=4, legend_label='qc pass', color='green', alpha=0.5)
    p1.circle(time, qc_suspect, size=4, legend_label='qc suspect', color='orange', alpha=0.7)
    p1.circle(time, qc_fail, size=6, legend_label='qc fail', color='red', alpha=1.0)
    p1.circle(time, qc_notrun, size=6, legend_label='qc not eval', color='gray', alpha=1.0)

    return p1


def bokeh_multi_plot(stream, results, title, **kwargs):

    kwargs = {
        **{
            'merge_tools': True,
            'toolbar_location': 'below',
            'sizing_mode': 'scale_width',
            'plot_width': 600,
            'plot_height': 200,
            'ncols': 2
        },
        **kwargs
    }

    plots = list(bokeh_multi_var(stream, results, title))
    return gridplot(plots, **kwargs)


def bokeh_multi_var(stream, results, title):
    for vname, qcobj in results.items():
        for modu, tests in qcobj.items():
            for testname, testresults in tests.items():
                plt = bokeh_plot_var(stream.time(), stream.data(vname), vname, qcobj, title, modu, testname)
                yield plt


def bokeh_plot_collected_results(results, **kwargs):
    kwargs = {
        **{
            'merge_tools': True,
            'toolbar_location': 'below',
            'sizing_mode': 'scale_width',
            'plot_width': 600,
            'plot_height': 200,
            'ncols': 2
        },
        **kwargs
    }

    plots = []
    for r in results:
        if r.data.any() and r.results.any():
            plots.append(bokeh_plot_collected_result(r))
    return gridplot(plots, **kwargs)


def bokeh_plot_collected_result(cr):
    title = f'{cr.stream_id}: {cr.package}.{cr.test}'
    p1 = plotting.figure(x_axis_type="datetime", title=title)
    p1.grid.grid_line_alpha = 0.3
    p1.xaxis.axis_label = 'Time'
    p1.yaxis.axis_label = 'Data'
    qc_pass = np.ma.masked_where(cr.results != 1, cr.data)
    qc_suspect = np.ma.masked_where(cr.results != 3, cr.data)
    qc_fail = np.ma.masked_where(cr.results != 4, cr.data)
    qc_notrun = np.ma.masked_where(cr.results != 2, cr.data)

    p1.line(cr.tinp, cr.data,  legend_label='data', color='#A6CEE3')
    p1.circle(cr.tinp, qc_notrun, size=3, legend_label='qc not run', color='gray', alpha=0.2)
    p1.circle(cr.tinp, qc_pass, size=4, legend_label='qc pass', color='green', alpha=0.5)
    p1.circle(cr.tinp, qc_suspect, size=4, legend_label='qc suspect', color='orange', alpha=0.7)
    p1.circle(cr.tinp, qc_fail, size=6, legend_label='qc fail', color='red', alpha=1.0)
    p1.circle(cr.tinp, qc_notrun, size=3, legend_label='qc not eval', color='gray', alpha=1.0)

    return p1



#!/usr/bin/env python
# coding=utf-8
"""Tests based on the IOOS QARTOD manuals."""
import logging
import warnings
from typing import Dict, List, Tuple, Union, Sequence
from numbers import Real as N
from collections import namedtuple

import numpy as np
import pandas as pd

from ioos_qc.utils import (
    isnan,
    isfixedlength,
    add_flag_metadata,
    great_circle_distance,
    mapdates
)


L = logging.getLogger(__name__)  # noqa


class QartodFlags(object):
    """Primary flags for QARTOD."""

    GOOD = 1
    UNKNOWN = 2
    SUSPECT = 3
    FAIL = 4
    MISSING = 9


FLAGS = QartodFlags  # Default name for all check modules
NOTEVAL_VALUE = QartodFlags.UNKNOWN

span = namedtuple('Span', 'minv maxv')


@add_flag_metadata(standard_name='aggregate_quality_flag',
                   long_name='Aggregate Quality Flag',
                   aggregate=True)
def aggregate(results: List) -> np.ma.MaskedArray:
    """
    Runs qartod_compare against all other qartod tests in results.
    """
    all_tests = [ r.results for r in results ]
    return qartod_compare(all_tests)


def qartod_compare(vectors : Sequence[Sequence[N]]
                   ) -> np.ma.MaskedArray:
    """Aggregates an array of flags by precedence into a single array.

    Args:
        vectors: An array of uniform length arrays representing individual flags

    Returns:
        A masked array of aggregated flag data.
    """
    shapes = [v.shape[0] for v in vectors]
    # Assert that all of the vectors are the same size.
    assert all([s == shapes[0] for s in shapes])
    assert all([v.ndim == 1 for v in vectors])

    result = np.ma.empty(shapes[0])
    result.fill(QartodFlags.MISSING)

    priorities = [
        QartodFlags.MISSING,
        QartodFlags.UNKNOWN,
        QartodFlags.GOOD,
        QartodFlags.SUSPECT,
        QartodFlags.FAIL
    ]
    # For each of the priorities in order, set the resultant array to the the
    # flag where that flag exists in each of the vectors.
    for p in priorities:
        for v in vectors:
            idx = np.where(v == p)[0]
            result[idx] = p
    return result.astype('uint8')


@add_flag_metadata(standard_name='location_test_quality_flag',
                   long_name='Location Test Quality Flag')
def location_test(lon : Sequence[N],
                  lat : Sequence[N],
                  bbox : Tuple[N, N, N, N] = (-180, -90, 180, 90),
                  range_max : N = None
                  ) -> np.ma.core.MaskedArray:
    """Checks that a location is within reasonable bounds.

    Checks that longitude and latitude are within reasonable bounds defaulting
    to lon = [-180, 180] and lat = [-90, 90].  Optionally, check for a maximum
    range parameter in great circle distance defaulting to meters which can
    also use a unit from the quantities library. Missing and masked data is
    flagged as UNKNOWN.

    Args:
        lon: Longitudes as a numeric numpy array or a list of numbers.
        lat: Latitudes as a numeric numpy array or a list of numbers.
        bbox: A length 4 tuple expressed in (minx, miny, maxx, maxy) [optional].
        range_max: Maximum allowed range expressed in geodesic curve distance (meters).

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    bboxnt = namedtuple('BBOX', 'minx miny maxx maxy')
    if bbox is not None:
        assert isfixedlength(bbox, 4)
        bbox = bboxnt(*bbox)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        lat = np.ma.masked_invalid(np.array(lat).astype(np.float64))
        lon = np.ma.masked_invalid(np.array(lon).astype(np.float64))

    if lon.shape != lat.shape:
        raise ValueError(
            'Lon ({0.shape}) and lat ({1.shape}) are different shapes'.format(
                lon, lat
            )
        )

    # Save original shape
    original_shape = lon.shape
    lon = lon.flatten()
    lat = lat.flatten()

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(lon.size, dtype='uint8')

    # If either lon or lat are masked we just set the flag to MISSING
    mloc = lon.mask & lat.mask
    flag_arr[mloc] = QartodFlags.MISSING

    # If there is only one masked value fail the location test
    mismatch = lon.mask != lat.mask
    flag_arr[mismatch] = QartodFlags.FAIL

    if range_max is not None and lon.size > 1:
        # Calculating the great_distance between each point
        # Flag suspect any distance over range_max
        d = great_circle_distance(lat, lon)
        flag_arr[d > range_max] = QartodFlags.SUSPECT

    # Ignore warnings when comparing NaN values even though they are masked
    # https://github.com/numpy/numpy/blob/master/doc/release/1.8.0-notes.rst#runtime-warnings-when-comparing-nan-numbers
    with np.errstate(invalid='ignore'):
        flag_arr[(lon < bbox.minx) | (lat < bbox.miny) |
                 (lon > bbox.maxx) | (lat > bbox.maxy)] = QartodFlags.FAIL

    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='gross_range_test_quality_flag',
                   long_name='Gross Range Test Quality Flag')
def gross_range_test(inp : Sequence[N],
                     fail_span : Tuple[N, N],
                     suspect_span : Tuple[N, N] = None
                     ) -> np.ma.core.MaskedArray:
    """Checks that values are within reasonable range bounds.

    Given a 2-tuple of minimum/maximum values, flag data outside of the given
    range as FAIL data.  Optionally also flag data which falls outside of a user
    defined range as SUSPECT. Missing and masked data is flagged as UNKNOWN.

    Args:
        inp: Input data as a numeric numpy array or a list of numbers.
        fail_span: 2-tuple range which to flag outside data as FAIL.
        suspect_span: 2-tuple range which to flag outside data as SUSPECT. [optional]

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    assert isfixedlength(fail_span, 2)
    sspan = span(*sorted(fail_span))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()
    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype='uint8')

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    if suspect_span is not None:
        assert isfixedlength(suspect_span, 2)
        uspan = span(*sorted(suspect_span))
        if uspan.minv < sspan.minv or uspan.maxv > sspan.maxv:
            raise ValueError('Suspect {} must fall within the Fail {}'.format(
                uspan, sspan
            ))
        # Flag suspect outside of user span
        with np.errstate(invalid='ignore'):
            flag_arr[(inp < uspan.minv) | (inp > uspan.maxv)] = QartodFlags.SUSPECT

    # Flag suspect outside of sensor span
    with np.errstate(invalid='ignore'):
        flag_arr[(inp < sspan.minv) | (inp > sspan.maxv)] = QartodFlags.FAIL

    return flag_arr.reshape(original_shape)


class ClimatologyConfig(object):
    """Objects to hold the config for a Climatology test

    Args:
        tspan: 2-tuple range.
               If period is defined, then this is a numeric range.
               If period is not defined, then its a date range.
        fspan: (optional) 2-tuple range of valid values. This is passed in as the fail_span to the gross_range_test.
        vspan: 2-tuple range of valid values. This is passed in as the suspect_span to the gross_range test.
        zspan: (optional) Vertical (depth) range, in meters positive down
        period: (optional) The unit the tspan argument is in. Defaults to datetime object
                but can also be any attribute supported by a pandas Timestamp object.

                See: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html

                Options:
                * year
                * week / weekofyear
                * dayofyear
                * dayofweek
                * quarter
    """
    mem = namedtuple('window', [
        'tspan',
        'fspan',
        'vspan',
        'zspan',
        'period'
    ])

    def __init__(self, members=None):
        members = members or []
        self._members = members

    @property
    def members(self):
        return self._members

    def values(self, tind : pd.Timestamp, zind=None):
        """
        Args:
            tind: Value to test for inclusion between time bounds
        """
        span = (None, None)
        for m in self._members:

            if m.period is not None:
                # If a period is defined, extract the attribute from the
                # pd.Timestamp object before comparison. The min and max
                # values are in this period unit already.
                tind_copy = getattr(tind, m.period)
            else:
                # If a period isn't defined, make a new Timestamp object
                # to align with the above name 'tind_copy'
                tind_copy = tind

            # If we are between times
            if tind_copy > m.tspan.minv and tind_copy <= m.tspan.maxv:
                if not isnan(zind) and not isnan(m.zspan):
                    # If we are between depths
                    if zind > m.zspan.minv and zind <= m.zspan.maxv:
                        span = m.vspan
                elif isnan(zind) and isnan(m.zspan):
                    span = m.vspan
        return span

    def add(self,
            tspan : Tuple[N, N],
            vspan : Tuple[N, N],
            fspan : Tuple[N, N] = None,
            zspan : Tuple[N, N] = None,
            period : str = None
            ) -> None:

        assert isfixedlength(tspan, 2)
        # If period is defined, tspan is a numeric
        # if it isn't defined, its a parsable date
        if period is not None:
            tspan = span(*sorted(tspan))
        else:
            tspan = span(*sorted([
                pd.Timestamp(tspan[0]),
                pd.Timestamp(tspan[1])
            ]))

        assert isfixedlength(vspan, 2)
        vspan = span(*sorted(vspan))

        if fspan is not None:
            assert isfixedlength(fspan, 2)
            fspan = span(*sorted(fspan))

        if zspan is not None:
            assert isfixedlength(zspan, 2)
            zspan = span(*sorted(zspan))

        if period is not None:
            # Test to make sure this is callable on a Timestamp
            try:
                getattr(pd.Timestamp.now(), period)
            except AttributeError:
                raise ValueError('The period "{period}" is not recognized')

        self._members.append(
            self.mem(
                tspan,
                fspan,
                vspan,
                zspan,
                period
            )
        )

    def check(self, tinp, inp, zinp):

        # Start with everything as UNKNOWN (2)
        flag_arr = np.ma.empty(inp.size, dtype='uint8')
        flag_arr.fill(QartodFlags.UNKNOWN)

        # If the value is masked set the flag to MISSING
        flag_arr[inp.mask] = QartodFlags.MISSING

        # Iterate over each member and apply its spans on the input data.
        # Member spans are applied in order and any data points that fall into
        # more than one member are flagged by each one.
        for m in self._members:

            if m.period is not None:
                # If a period is defined, extract the attribute from the
                # pd.DatetimeIndex object before comparison. The min and max
                # values are in this period unit already.
                tinp_copy = getattr(tinp, m.period).to_series()
            else:
                # If a period isn't defined, make a new Timestamp object
                # to align with the above name 'tinp_copy'
                tinp_copy = tinp

            # If a zspan is defined but we don't have z input (zinp), skip this member
            # Note: `zinp.count()` can return `np.ma.masked` so we also check using isnan
            if not isnan(m.zspan) and (not zinp.count() or isnan(zinp.any())):
                continue

            # Indexes that align with the T
            t_idx = (tinp_copy >= m.tspan.minv) & (tinp_copy <= m.tspan.maxv)

            # Indexes that align with the Z
            if not isnan(m.zspan):
                # Only test non-masked values between the min and max.
                # Ignore warnings about comparing masked values
                with np.errstate(invalid='ignore'):
                    z_idx = (~zinp.mask) & (zinp >= m.zspan.minv) & (zinp <= m.zspan.maxv)
            else:
                # Only test the values with masked Z, ie values with no Z
                z_idx = zinp.mask

            # Combine the T and Z indexes
            values_idx = (t_idx & z_idx)

            # Failed and suspect data for this value span. Combining fail_idx or
            # suspect_idx with values_idx represents the subsets of data that should be
            # fail and suspect respectively.
            if not isnan(m.fspan):
                fail_idx = (inp < m.fspan.minv) | (inp > m.fspan.maxv)
            else:
                fail_idx = np.zeros(inp.size, dtype=bool)

            suspect_idx = (inp < m.vspan.minv) | (inp > m.vspan.maxv)

            with np.errstate(invalid='ignore'):
                flag_arr[(values_idx & fail_idx)] = QartodFlags.FAIL
                flag_arr[(values_idx & ~fail_idx & suspect_idx)] = QartodFlags.SUSPECT
                flag_arr[(values_idx & ~fail_idx & ~suspect_idx)] = QartodFlags.GOOD

        return flag_arr

    @staticmethod
    def convert(config):
        # Create a ClimatologyConfig object if one was not passed in
        if isinstance(config, ClimatologyConfig):
            return config

        c = ClimatologyConfig()
        for climate_config_dict in config:
            c.add(**climate_config_dict)
        return c


@add_flag_metadata(standard_name='climatology_test_quality_flag',
                   long_name='Climatology Test Quality Flag')
def climatology_test(config : Union[ClimatologyConfig, Sequence[Dict[str, Tuple]]],
                     inp : Sequence[N],
                     tinp : Sequence[N],
                     zinp : Sequence[N],
                     ) -> np.ma.core.MaskedArray:
    """Checks that values are within reasonable range bounds and flags as SUSPECT.

    Data for which no ClimatologyConfig member exists is marked as UNKNOWN.

    Args:
        config: A ClimatologyConfig object or a list of dicts containing tuples
            that can be used to create a ClimatologyConfig object. See ClimatologyConfig
            docs for more info.
        tinp: Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
          This includes numpy datetime64, python datetime objects and pandas Timestamp object.
          ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()]
          If anything else is passed in the format is assumed to be seconds since the unix epoch.
        vinp: Input data as a numeric numpy array or a list of numbers.
        zinp: Z (depth) data, in meters positive down, as a numeric numpy array or a list of numbers.

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    # Create a ClimatologyConfig object if one was not passed in
    config = ClimatologyConfig.convert(config)

    tinp = mapdates(tinp)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))
        zinp = np.ma.masked_invalid(np.array(zinp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape

    # We compare using a pandas Timestamp for helper functions like
    # 'week' and 'dayofyear'. It is surprisingly hard to pull these out
    # of a plain datetime64 object.
    tinp = pd.DatetimeIndex(tinp.flatten())
    inp = inp.flatten()
    zinp = zinp.flatten()

    flag_arr = config.check(tinp, inp, zinp)
    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='spike_test_quality_flag',
                   long_name='Spike Test Quality Flag')
def spike_test(inp : Sequence[N],
               suspect_threshold: N,
               fail_threshold: N
               ) -> np.ma.core.MaskedArray:
    """Check for spikes by checking neighboring data against thresholds

    Determine if there is a spike at data point n-1 by subtracting
    the midpoint of n and n-2 and taking the absolute value of this
    quantity, and checking if it exceeds a low or high threshold.
    Values which do not exceed either threshold are flagged GOOD,
    values which exceed the low threshold are flagged SUSPECT,
    and values which exceed the high threshold are flagged FAIL.
    Missing and masked data is flagged as UNKNOWN.

    Args:
        inp: Input data as a numeric numpy array or a list of numbers.
        suspect_threshold: The SUSPECT threshold value, in observations units.
        fail_threshold: The SUSPECT threshold value, in observations units.

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Calculate the average of n-2 and n
    ref = np.zeros(inp.size, dtype=np.float64)
    ref[1:-1] = (inp[0:-2] + inp[2:]) / 2
    ref = np.ma.masked_invalid(ref)

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype='uint8')

    # Calculate the (n-1 - ref) difference
    diff = np.abs(inp - ref)

    # If n-1 - ref is greater than the low threshold, SUSPECT test
    with np.errstate(invalid='ignore'):
        flag_arr[diff > suspect_threshold] = QartodFlags.SUSPECT

    # If n-1 - ref is greater than the high threshold, FAIL test
    with np.errstate(invalid='ignore'):
        flag_arr[diff > fail_threshold] = QartodFlags.FAIL

    # test is undefined for first and last values
    flag_arr[0] = QartodFlags.UNKNOWN
    flag_arr[-1] = QartodFlags.UNKNOWN

    # If the value is masked or nan set the flag to MISSING
    flag_arr[diff.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='rate_of_change_test_quality_flag',
                   long_name='Rate of Change Test Quality Flag')
def rate_of_change_test(inp : Sequence[N],
                        tinp : Sequence[N],
                        threshold : float
                        ) -> np.ma.core.MaskedArray:
    """Checks the first order difference of a series of values to see if
    there are any values exceeding a threshold defined by the inputs.
    These are then marked as SUSPECT.  It is up to the test operator
    to determine an appropriate threshold value for the absolute difference not to
    exceed. Threshold is expressed as a rate in observations units per second.
    Missing and masked data is flagged as UNKNOWN.

    Args:
        inp: Input data as a numeric numpy array or a list of numbers.
        tinp: Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
              This includes numpy datetime64, python datetime objects and pandas Timestamp object.
              ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()]
              If anything else is passed in the format is assumed to be seconds since the unix epoch.
        threshold: A float value representing a rate of change over time,
                   in observation units per second.

    Returns:
        A masked array of flag values equal in size to that of the input.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype='uint8')

    # calculate rate of change in units/second
    roc = np.ma.zeros(inp.size, dtype='float')

    tinp = mapdates(tinp).flatten()
    roc[1:] = np.abs(np.diff(inp) / np.diff(tinp).astype('timedelta64[s]').astype(float))

    with np.errstate(invalid='ignore'):
        flag_arr[roc > threshold] = QartodFlags.SUSPECT

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='flat_line_test_quality_flag',
                   long_name='Flat Line Test Quality Flag')
def flat_line_test(inp: Sequence[N],
                   tinp: Sequence[N],
                   suspect_threshold: int,
                   fail_threshold: int,
                   tolerance: N = 0
                   ) -> np.ma.MaskedArray:
    """Check for consecutively repeated values within a tolerance.
    Missing and masked data is flagged as UNKNOWN.
    More information: https://github.com/ioos/ioos_qc/pull/11

    Args:
        inp: Input data as a numeric numpy array or a list of numbers.
        tinp: Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
              This includes numpy datetime64, python datetime objects and pandas Timestamp object.
              ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()]
              If anything else is passed in the format is assumed to be seconds since the unix epoch.
        suspect_threshold: The number of seconds within `tolerance` to
            allow before being flagged as SUSPECT.
        fail_threshold: The number of seconds within `tolerance` to
            allow before being flagged as FAIL.
        tolerance: The tolerance that should be exceeded between consecutive values.
            To determine if the current point `n` should be flagged, we use a rolling window, with endpoint at
            point `n`, and calculate the range of values in the window. If that range is less than `tolerance`,
            then the point is flagged.

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    # input as numpy arr
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Start with everything as passing
    flag_arr = np.full((inp.size,), QartodFlags.GOOD)

    # if we have fewer than 3 points, we can't run the test, so everything passes
    if len(inp) < 3:
        return flag_arr.reshape(original_shape)

    # determine median time interval
    tinp = mapdates(tinp).flatten()

    # The thresholds are in seconds so we round make sure the interval is also in seconds
    time_interval = np.median(np.diff(tinp)).astype('timedelta64[s]').astype(float)

    def rolling_window(a, window):
        """
        https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html
        """
        if len(a) < window:
            return np.ma.MaskedArray(np.empty((0, window + 1)))
        shape = a.shape[:-1] + (a.shape[-1] - window + 1, window + 1)
        strides = a.strides + (a.strides[-1],)
        arr = np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)
        return np.ma.masked_invalid(arr[:-1, :])

    def run_test(test_threshold, flag_value):
        # convert time thresholds to number of observations
        count = (int(test_threshold) / time_interval).astype(int)

        # calculate actual data ranges for each window
        window = rolling_window(inp, count)
        data_min = np.min(window, 1)
        data_max = np.max(window, 1)
        data_range = np.abs(data_max - data_min)

        # find data ranges that are within threshold and flag them
        test_results = np.ma.filled(data_range < tolerance, fill_value=False)
        # data points before end of first window should pass
        n_fill = count if count < len(inp) else len(inp)
        test_results = np.insert(test_results, 0, np.full((n_fill,), False))
        flag_arr[test_results] = flag_value

    run_test(suspect_threshold, QartodFlags.SUSPECT)
    run_test(fail_threshold, QartodFlags.FAIL)

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='attenuated_signal_test_quality_flag',
                   long_name='Attenuated Signal Test Quality Flag')
def attenuated_signal_test(inp : Sequence[N],
                           tinp : Sequence[N],
                           suspect_threshold: N,
                           fail_threshold: N,
                           test_period: N = None,
                           min_obs: N = None,
                           min_period: int = None,
                           check_type : str = 'std',
                           *args,
                           **kwargs,
                           ) -> np.ma.MaskedArray:
    """Check for near-flat-line conditions using a range or standard deviation.

    Missing and masked data is flagged as UNKNOWN.

    Args:
        inp: Input data as a numeric numpy array or a list of numbers.
        tinp: Time input data as a numpy array of dtype `datetime64`.
        suspect_threshold: Any calculated value below this amount will be flagged as SUSPECT.
            In observations units.
        fail_threshold: Any calculated values below this amount will be flagged as FAIL.
            In observations units.
        test_period: Length of time to test over in seconds [optional].
            Otherwise, will test against entire `inp`.
        min_obs: Minimum number of observations in window required to calculate a result [optional].
            Otherwise, test will start at beginning of time series.
            Note: you can specify either `min_obs` or `min_period`, but not both.
        min_period: Minimum number of seconds in test_period required to calculate a result [optional].
            Otherwise, test will start at beginning of time series.
            Note: you can specify either `min_obs` or `min_period`, but not both.
        check_type: Either 'std' (default) or 'range', depending on the type of check
            you wish to perform.

    Returns:
        A masked array of flag values equal in size to that of the input.
        This array will always contain only a single unique value since all
        input data is flagged together.
    """

    # window_func: Applied to each window when `time_period` is supplied
    # check_func: Applied to a flattened numpy array when no `time_period` is supplied
    # These are split for performance reasons
    if check_type == 'std':
        window_func = lambda x: x.std()  # noqa
        check_func = np.std
    elif check_type == 'range':
        def window_func(w):
            # When pandas>=1.0 and numba are installed, this is about twice as fast
            try:
                return w.apply(np.ptp, raw=True, engine='numba')
            except (ImportError, TypeError):
                return w.apply(np.ptp, raw=True)
        check_func = np.ptp
    else:
        raise ValueError('Check type "{}" is not one of ["std", "range"]'.format(check_type))

    tinp = mapdates(tinp)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape

    # Start with everything as not tested (0)
    flag_arr = np.full((inp.size,), QartodFlags.UNKNOWN)

    if test_period:
        if min_obs is not None:
            min_periods = min_obs
        elif min_period is not None:
            time_interval = np.median(np.diff(tinp)).astype('timedelta64[s]').astype(float)
            min_periods = (min_period / time_interval).astype(int)
        else:
            min_periods = None
        series = pd.Series(inp.flatten(), index=tinp.flatten())
        windows = series.rolling(f'{test_period}s', min_periods=min_periods)
        check_val = window_func(windows)
    else:
        # applying np.ptp to Series causes warnings, this is a workaround
        series = inp.flatten()
        check_val = np.ones_like(flag_arr) * check_func(series)

    flag_arr[check_val >= suspect_threshold] = QartodFlags.GOOD
    flag_arr[check_val < suspect_threshold] = QartodFlags.SUSPECT
    flag_arr[np.isnan(check_val)] = QartodFlags.UNKNOWN
    flag_arr[check_val < fail_threshold] = QartodFlags.FAIL
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(standard_name='density_inversion_test_flag',
                   long_name='Density Inversion Test Flag')
def density_inversion_test(inp: Sequence[N],
                           zinp: Sequence[N],
                           suspect_threshold: float = None,
                           fail_threshold: float = None
                           ) -> np.ma.core.MaskedArray:
    """
    With few exceptions, potential water density will increase with increasing pressure. When
    vertical profile data is obtained, this test is used to flag as failed T, C, and SP observations, which
    yield densities that do not sufficiently increase with pressure. A small operator-selected density
    threshold (DT) allows for micro-turbulent exceptions. This test can be run on downcasts, upcasts,
    or down/up cast results produced in real time.

    Both Temperature and Salinity should be flagged based on the result of this test.

    Ref: Manual for Real-Time Quality Control of in-situ Temperature and Salinity Data, Version 2.0, January 2016

    Args:
        inp: Potential density values as a numeric numpy array or a list of numbers.
        zinp: Corresponding depth/pressure values for each density.
        suspect_threshold: A float value representing a maximum potential density(or sigma0)
            variation to be tolerated, downward density variation exceeding this will be flagged as SUSPECT.
        fail_threshold:  A float value representing a maximum potential density(or sigma0)
            variation to be tolerated, downward density variation exceeding this will be flagged as FAIL.
    Returns:
        A masked array of flag values equal in size to that of the input.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))
        zinp = np.ma.masked_invalid(np.array(zinp).astype(np.float64))

    # Make sure both inputs are the same size.
    if inp.shape != zinp.shape:
        raise ValueError(f'Density ({inp.shape}) and depth ({zinp.shape}) must be the same shape')

    # Start with everything as passing
    flag_arr = QartodFlags.GOOD * np.ma.ones(inp.size, dtype='uint8')

    # If no data or just one record, return respectively an empty mask array or UNKNOWN
    if inp.size == 0:
        return np.ma.masked_array([])
    if inp.size < 2:
        flag_arr[0] = QartodFlags.UNKNOWN
        return flag_arr

    # Compute the vertical density variability along zinp and flip delta according to zinp variation direction
    delta = np.sign(np.diff(zinp)) * np.diff(inp)

    if suspect_threshold is not None:
        with np.errstate(invalid='ignore'):
            is_suspect = delta < suspect_threshold
            if any(is_suspect):
                flag_arr[:-1][is_suspect == True] = QartodFlags.SUSPECT  # noqa:E712- Previous value
                flag_arr[1:][is_suspect == True] = QartodFlags.SUSPECT  # noqa:E712- Reversed value

    if fail_threshold is not None:
        with np.errstate(invalid='ignore'):
            is_fail = delta < fail_threshold
            if any(is_fail):
                flag_arr[:-1][is_fail == True] = QartodFlags.FAIL  # noqa:E712- Previous value
                flag_arr[1:][is_fail == True] = QartodFlags.FAIL  # noqa:E712- Reversed Value

    # If the value or depth is masked set the flag to MISSING for this record and the following one.
    is_missing = inp.mask | zinp.mask
    flag_arr[is_missing] = QartodFlags.MISSING
    flag_arr[1:][is_missing[:-1]] = QartodFlags.MISSING
    return flag_arr



#!/usr/bin/env python
# coding=utf-8
"""Tests based on the ARGO QC manual."""
import logging
import warnings
from numbers import Real as N
from typing import Sequence

import numpy as np

from ioos_qc.qartod import QartodFlags
from ioos_qc.utils import add_flag_metadata
from ioos_qc.utils import great_circle_distance
from ioos_qc.utils import mapdates

L = logging.getLogger(__name__)  # noqa


@add_flag_metadata(stanard_name='pressure_increasing_test_quality_flag',
                   long_name='Pressure Increasing Test Quality Flag')
def pressure_increasing_test(inp):
    """
    Returns an array of flag values where each input is flagged with SUSPECT if
    it does not monotonically increase

    Ref: ARGO QC Manual: 8. Pressure increasing test

    Args:
        inp: Pressure values as a numeric numpy array or a list of numbers.
    Returns:
        A masked array of flag values equal in size to that of the input.
    """
    delta = np.diff(inp)
    flags = np.ones_like(inp, dtype='uint8') * QartodFlags.GOOD

    # Correct for downcast vs upcast by flipping the sign if it's decreasing
    sign = np.sign(np.mean(delta))
    if sign < 0:
        delta = sign * delta

    flag_idx = np.where(delta <= 0)[0] + 1
    flags[flag_idx] = QartodFlags.SUSPECT

    return flags


@add_flag_metadata(standard_name='speed_test_quality_flag',
                   long_name='Speed Test Quality Flag')
def speed_test(lon: Sequence[N],
               lat: Sequence[N],
               tinp: Sequence[N],
               suspect_threshold: float,
               fail_threshold: float
               ) -> np.ma.core.MaskedArray:
    """Checks that the calculated speed between two points is within reasonable bounds.

    This test calculates a speed between subsequent points by
      * using latitude and longitude to calculate the distance between points
      * calculating the time difference between those points
      * checking if distance/time_diff exceeds the given threshold(s)

    Missing and masked data is flagged as UNKNOWN.

    If this test fails, it typically means that either a position or time is bad data,
     or that a platform is mislabeled.

    Ref: ARGO QC Manual: 5. Impossible speed test

    Args:
        lon: Longitudes as a numeric numpy array or a list of numbers.
        lat: Latitudes as a numeric numpy array or a list of numbers.
        tinp: Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
              This includes numpy datetime64, python datetime objects and pandas Timestamp object.
              ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()]
              If anything else is passed in the format is assumed to be seconds since the unix epoch.
        suspect_threshold: A float value representing a speed, in meters per second.
           Speeds exceeding this will be flagged as SUSPECT.
        fail_threshold: A float value representing a speed, in meters per second.
           Speeds exceeding this will be flagged as FAIL.

    Returns:
        A masked array of flag values equal in size to that of the input.
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        lat = np.ma.masked_invalid(np.array(lat).astype(np.float64))
        lon = np.ma.masked_invalid(np.array(lon).astype(np.float64))
        tinp = mapdates(tinp)

    if lon.shape != lat.shape or lon.shape != tinp.shape:
        raise ValueError(f'Lon ({lon.shape}) and lat ({lat.shape}) and tinp ({tinp.shape}) must be the same shape')

    # Save original shape
    original_shape = lon.shape
    lon = lon.flatten()
    lat = lat.flatten()
    tinp = tinp.flatten()

    # If no data, return
    if lon.size == 0:
        return np.ma.masked_array([])

    # Start with everything as passing
    flag_arr = QartodFlags.GOOD * np.ma.ones(lon.size, dtype='uint8')

    # If either lon or lat are masked we just set the flag to MISSING
    mloc = lon.mask & lat.mask
    flag_arr[mloc] = QartodFlags.MISSING

    # If only one data point, return
    if lon.size < 2:
        flag_arr[0] = QartodFlags.UNKNOWN
        return flag_arr.reshape(original_shape)

    # Calculate the great_distance between each point
    dist = great_circle_distance(lat, lon)

    # calculate speed in m/s
    speed = np.ma.zeros(tinp.size, dtype='float')
    speed[1:] = np.abs(dist[1:] / np.diff(tinp).astype('timedelta64[s]').astype(float))

    with np.errstate(invalid='ignore'):
        flag_arr[speed > suspect_threshold] = QartodFlags.SUSPECT

    with np.errstate(invalid='ignore'):
        flag_arr[speed > fail_threshold] = QartodFlags.FAIL

    # first value is unknown, since we have no speed data for the first point
    flag_arr[0] = QartodFlags.UNKNOWN

    # If the value is masked set the flag to MISSING
    flag_arr[dist.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)



from .config_creator import (
    CreatorConfig,
    QcConfigCreator,
    QcVariableConfig,
    QC_CONFIG_CREATOR_SCHEMA,
    VARIABLE_CONFIG_SCHEMA
)

__all__ = [
    'CreatorConfig',
    'QcConfigCreator',
    'QcVariableConfig',
    'QC_CONFIG_CREATOR_SCHEMA',
    'VARIABLE_CONFIG_SCHEMA'
]



# module pyparsing.py
#
# Copyright (c) 2003-2019  Paul T. McGuire
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# flake8: noqa
from pyparsing import (
    Literal,
    Word,
    Group,
    Forward,
    alphas,
    alphanums,
    Regex,
    CaselessKeyword,
    Suppress,
    delimitedList,
)
import math
import operator

# map operator symbols to corresponding arithmetic operations
epsilon = 1e-12
opn = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "^": operator.pow,
}

fn = {
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "exp": math.exp,
    "abs": abs,
    "trunc": lambda a: int(a),
    "round": round,
    "sgn": lambda a: -1 if a < -epsilon else 1 if a > epsilon else 0,
}

exprStack = []


def push_first(toks):
    exprStack.append(toks[0])


def push_unary_minus(toks):
    for t in toks:
        if t == "-":
            exprStack.append("unary -")
        else:
            break


def BNF():
    """
    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*
    """
    # use CaselessKeyword for e and pi, to avoid accidentally matching
    # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
    # and CaselessKeyword only match whole words
    e = CaselessKeyword("E")
    pi = CaselessKeyword("PI")
    # fnumber = Combine(Word("+-"+nums, nums) +
    #                    Optional("." + Optional(Word(nums))) +
    #                    Optional(e + Word("+-"+nums, nums)))
    # or use provided pyparsing_common.number, but convert back to str:
    # fnumber = ppc.number().addParseAction(lambda t: str(t[0]))
    fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
    ident = Word(alphas, alphanums + "_$")

    plus, minus, mult, div = map(Literal, "+-*/")
    lpar, rpar = map(Suppress, "()")
    addop = plus | minus
    multop = mult | div
    expop = Literal("^")

    expr = Forward()
    expr_list = delimitedList(Group(expr))
    # add parse action that replaces the function identifier with a (name, number of args) tuple
    fn_call = (ident + lpar - Group(expr_list) + rpar).setParseAction(
        lambda t: t.insert(0, (t.pop(0), len(t[0])))
    )
    atom = (
        addop[...]
        + (
            (fn_call | pi | e | fnumber | ident).setParseAction(push_first)
            | Group(lpar + expr + rpar)
        )
    ).setParseAction(push_unary_minus)

    # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
    # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
    factor = Forward()
    factor <<= atom + (expop + factor).setParseAction(push_first)[...]
    term = factor + (multop + factor).setParseAction(push_first)[...]
    expr <<= term + (addop + term).setParseAction(push_first)[...]
    bnf = expr

    return bnf


def evaluate_stack(s, stats):
    op, num_args = s.pop(), 0
    if isinstance(op, tuple):
        op, num_args = op
    if op == "unary -":
        return -evaluate_stack(s, stats)
    if op in "+-*/^":
        # note: operands are pushed onto the stack in reverse order
        op2 = evaluate_stack(s, stats)
        op1 = evaluate_stack(s, stats)
        return opn[op](op1, op2)
    elif op == "PI":
        return math.pi  # 3.1415926535
    elif op == "E":
        return math.e  # 2.718281828
    elif op == "mean":
        return stats['mean']
    elif op == "min":
        return stats['min']
    elif op == "max":
        return stats['max']
    elif op == "std":
        return stats['std']
    elif op in fn:
        # note: args are pushed onto the stack in reverse order
        args = reversed([evaluate_stack(s, stats) for _ in range(num_args)])
        return fn[op](*args)
    elif op[0].isalpha():
        raise Exception("invalid identifier '%s'" % op)
    else:
        return float(op)


def eval_fx(fx, stats):
    """Given fx and stats ('min', 'max', 'mean', 'std') return the result"""
    _ = BNF().parseString(fx, parseAll=True)
    val = evaluate_stack(exprStack[:], stats)

    return val
